-------------------------------------------------------------------------------
-- Carrier sense CRS not used, because full duplex only
-- FULL DUPLEX only and NO flow control
-- designed to allow jumbo frame 
-- O. Bourrion 12/2013
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use work.pack_mac.all;

entity mac_tx_engine is
port (
	rstn : in std_logic;
	
	-- to user interface
	clk : in std_logic;
	tx_tdata : in std_logic_vector(7 downto 0); --!frame data to be Xfrerd
	tx_tvalid : in std_logic; --! TX data valid
	tx_tlast : in std_logic; --! final Xfer of frame
	tx_tuser : in std_logic; --! Signal tx error
	tx_tready : out std_logic; --! Transmitter ready, qualifies when data is taken by MAC
	tx_statistics_vector : out std_logic_vector(31 downto 0); 
	tx_statistics_valid : out std_logic;
	tx_mac_config_vector : in std_logic_vector(79 downto 0);

	-- to gmii
	gmii_txd : out std_logic_vector(7 downto 0);
	gmii_tx_en : out std_logic;
	gmii_tx_er : out std_logic
	);
end mac_tx_engine;

architecture RTL of mac_tx_engine is
component crc_gen
generic (is_check : boolean :=false);
port(
	rstn : in std_logic;
	clk : in std_logic;
	clear_crc : in std_logic;

	data_en : in std_logic;
	data_in : in std_logic_vector(7 downto 0);

	crc_is_good : out std_logic;
	checksum: out std_logic_vector(7 downto 0)
);
end component;

type def_state_tx is (idle,preamble,SFD,data_xfer,pad,FCS,IFG);
signal state_tx : def_state_tx;

type def_stretcher is (idle, pulse);
signal stretcher : def_stretcher;
signal stretch_cnt : natural range 0 to minFrameSize-2;
signal s_tx_ready_stretch : std_logic;

signal byte_cnt : natural range 0 to maxFrameSize-1;
signal preamble_cnt : natural range 0 to 6;

type def_delay_data is array (0 to 1) of std_logic_vector(7 downto 0);
signal data_saved : def_delay_data;
signal s_gmii_txd : std_logic_vector(7 downto 0);
signal checksum : std_logic_vector(7 downto 0);
signal s_tx_tdata : std_logic_vector(7 downto 0);

signal crc_en : std_logic;
signal clear_crc : std_logic;

signal s_tx_tready : std_logic;
-- signal s_tx_tready_1ck : std_logic;
signal s_tx_tlast_1ck : std_logic;
signal s_tx_tlast_2ck : std_logic;

begin
tx_statistics_vector<=(others=>'0');
tx_statistics_valid<='0';

tx_tready<=s_tx_tready;
process(rstn,clk)
begin
	if rstn='0' then
		s_tx_ready_stretch<='0';
		stretch_cnt<=0;
		stretcher<=idle;

	elsif rising_edge(clk) then
		case stretcher is
			when idle =>
				stretch_cnt<=0;
				if (state_tx=SFD) then
					s_tx_ready_stretch<='1';
					stretcher<=pulse;
				end if;
			when pulse =>
				if (stretch_cnt=minFrameSize-4-3) then
					s_tx_ready_stretch<='0';
					stretcher<=idle;
				else
					stretch_cnt<=stretch_cnt+1;
				end if;
			end case;
	end if;
end process;

process(rstn,clk)
begin
	if rstn='0' then
		byte_cnt<=0;
		preamble_cnt<=0;
		s_tx_tready<='0';
-- 		s_tx_tready_1ck<='0';
		clear_crc<='0';
		gmii_tx_er<='0';
		s_tx_tlast_1ck<='0';
		s_tx_tlast_2ck<='0';

		s_gmii_txd<=(others=>'0');
		data_saved<=(others=>(others=>'0'));
		gmii_tx_en<='0';

		state_tx<=idle;

	elsif rising_edge(clk) then
		gmii_tx_er<=tx_tuser;
-- 		s_tx_tready_1ck<=s_tx_tready;
		s_tx_tlast_1ck<=tx_tlast;
		s_tx_tlast_2ck<=s_tx_tlast_1ck;

		if (tx_tvalid='1' and s_tx_tready='1') then
			data_saved(0)<=tx_tdata;
			data_saved(1)<=data_saved(0);
		end if;

		case state_tx is
			when idle =>
				preamble_cnt<=0;
				clear_crc<='0';
				if tx_tvalid='1' then
					s_gmii_txd<=X"55";
					gmii_tx_en<='1';
					s_tx_tready<='1';
					state_tx<=preamble;
				else
					s_gmii_txd<=(others=>'0');
					gmii_tx_en<='0';
					s_tx_tready<='0';
				end if;

			when preamble =>
				byte_cnt<=0;
				if preamble_cnt=1 then
					s_tx_tready<='0';
				end if;

				if preamble_cnt=6 then
					s_gmii_txd<=X"D5";
					state_tx<=SFD;
				else
					preamble_cnt<=preamble_cnt+1;
				end if;

			when SFD =>
				byte_cnt<=byte_cnt+1;
				s_gmii_txd<=data_saved(1);
				s_tx_tready<='1';
				state_tx<=data_xfer;

			when data_xfer =>
				byte_cnt<=byte_cnt+1;
				if (tx_tlast='1') then
					s_tx_tready<='0';
				end if;

				if (s_tx_tlast_2ck='1') then
					if (byte_cnt>=minFrameSize-4) then --no padding
						byte_cnt<=0;
						s_gmii_txd<=checksum;
						state_tx<=FCS;
					else -- padding required
						s_gmii_txd<=(others=>'0');
						state_tx<=pad;
					end if;
				else
					s_gmii_txd<=data_saved(0);
				end if;

			when pad =>
				if (byte_cnt=minFrameSize-4-2) then
					s_tx_tready<='0';
				end if;
				if (byte_cnt=minFrameSize-4) then
					byte_cnt<=0;
					s_gmii_txd<=checksum;
					state_tx<=FCS;
				else
					s_gmii_txd<=(others=>'0');
					byte_cnt<=byte_cnt+1;
				end if;

			when FCS =>
				s_gmii_txd<=checksum;
				if (byte_cnt=3) then
					byte_cnt<=0;
					gmii_tx_en<='0';
					clear_crc<='1';
					state_tx<=IFG;
				else
					byte_cnt<=byte_cnt+1;
				end if;

			when IFG =>
				clear_crc<='0';
				s_gmii_txd<=(others=>'0');
				if (byte_cnt=interpacketgap-2) then
					state_tx<=idle;
				else
					byte_cnt<=byte_cnt+1;
				end if;
		end case;
	end if;
end process;
gmii_txd<=s_gmii_txd;

crc_en<=s_tx_tready or s_tx_ready_stretch;
s_tx_tdata<=tx_tdata when  s_tx_tready='1' else x"00";

lbl_crc : crc_gen generic map(is_check=>false)
	port map(
		rstn=>rstn,
		clk=>clk,
		clear_crc=>clear_crc,

		data_en=>crc_en,
		data_in=>s_tx_tdata,

		crc_is_good=>open,
		checksum=>checksum
		);

end RTL;

