-------------------------------------------------------------------------------
-- see http://www.tcpipguide.com/free/t_DHCPMessageFormat.htm
-- http://www.tcpipguide.com/free/t_DHCPOptionsOptionFormatandOptionOverloading-2.htm
-- http://www.tcpipguide.com/free/t_SummaryOfDHCPOptionsBOOTPVendorInformationFields-2.htm
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.pack_axi.all;
use work.pack_ipv4_types.all;
use work.pack_arp_types.all;

entity dhcp_client is
generic (
		CLOCK_FREQ		: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
		LEASE_TIME_TO_REQ	: integer := 3600 -- lease time request (in sec)
		);
port (
	-- system signals
	clk					: in  STD_LOGIC;
	reset 				: in  STD_LOGIC;
	our_mac_address 	: in std_logic_vector (47 downto 0);

	-- UDP TX signals
	udp_tx_start		: out std_logic;							-- indicates req to tx UDP
	udp_txi				: out udp_tx_type;							-- UDP tx cxns
	udp_tx_result		: in std_logic_vector (1 downto 0);-- tx status (changes during transmission)
	udp_tx_data_out_ready: in std_logic;							-- indicates udp_tx is ready to take data
	
	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- UDP RX signals
	udp_rx_start		: in std_logic;							-- indicates receipt of udp header
	udp_rxo				: in udp_rx_type;

	allocated_ip_address : out STD_LOGIC_VECTOR (31 downto 0)
);
end dhcp_client;


architecture RTL of dhcp_client is
type def_state_dhcp is (init,selecting,requesting,bound,renewing,rebinding);
signal state_dhcp : def_state_dhcp;
type def_dhcp_type is (DHCPDISCOVER,DHCPREQUEST);
signal dhcp_type : def_dhcp_type;
signal dhcp_type2xmit : def_dhcp_type;

constant lease_time_vect: std_logic_vector(31 downto 0):=std_logic_vector(to_unsigned(LEASE_TIME_TO_REQ,32));
signal client_ip_add_s : STD_LOGIC_VECTOR (31 downto 0);
signal offered_client_ip_add_s : STD_LOGIC_VECTOR (31 downto 0);
signal server_ip_add_s : STD_LOGIC_VECTOR (31 downto 0);
signal rcv_server_ip_add_s : STD_LOGIC_VECTOR (31 downto 0);
signal freq_scaler		: unsigned (31 downto 0);		-- scales data_in_clk downto 1Hz
signal timer_enable : std_logic;
signal init_timer : natural range 0 to 10;
signal timer : unsigned (31 downto 0);
signal secs : unsigned (15 downto 0);

signal do_tx_dhcp : std_logic;
signal rcv_dhcp_offer : std_logic;
signal rcv_dhcp_nack : std_logic;
signal rcv_dhcp_ack : std_logic;
signal lease_time_offered	: unsigned (31 downto 0);
signal T1	: unsigned (31 downto 0);
signal T2	: unsigned (31 downto 0);

constant nb_src : natural :=2;
signal src_udp_tx_start : std_logic_vector(nb_src-1 downto 0);
signal src_udp_tx : udp_tx_vect_type(nb_src-1 downto 0);

type def_state_dhcp_tx is (idle,set_len,xmit_bootp,xmit_dhcp_discover,xmit_dhcp_request,finish);
signal state_dhcp_tx : def_state_dhcp_tx;
signal tx_count : natural range 0 to 255;
signal rx_count : natural range 0 to 255;
signal option_len : unsigned(7 downto 0);
signal option: std_logic_vector(7 downto 0);
signal dhcp_message_type : std_logic_vector(7 downto 0);

type def_state_dhcp_rx is (idle,rcv_bootp,rcv_option_type,rcv_option_len,rcv_option_data,wait_end,finish);
signal state_dhcp_rx : def_state_dhcp_rx;
signal udp_tx_result_dhcp		: std_logic_vector (1 downto 0);
signal udp_tx_data_out_ready_dhcp: std_logic;

type def_state_arb is (idle,busy);
signal state_arb_udp : def_state_arb;

signal sel : natural range 0 to nb_src-1;
begin
-------------------------------------------------------------------------------
-- FREQ scaler
-------------------------------------------------------------------------------
allocated_ip_address<=client_ip_add_s;
process(clk)
begin
	if rising_edge(clk) then
		if reset = '1' then
			freq_scaler <= to_unsigned(CLOCK_FREQ,32);
			timer_enable<='0';
		else
			-- freq scaling and 1-sec timer
			if freq_scaler = x"00000000" then
				freq_scaler <= to_unsigned(CLOCK_FREQ,32);
				timer_enable<='1';
			else
				freq_scaler <= freq_scaler - 1;
				timer_enable<='0';
			end if;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- DHCP state machine
-------------------------------------------------------------------------------
process(clk)
begin
	if rising_edge(clk) then
		if reset='1' then
			secs<=(others=>'0');
			init_timer<=0;
			do_tx_dhcp<='0';
			state_dhcp<=init;
		else
			case state_dhcp is
				when init =>
					server_ip_add_s<=(others=>'1');
					client_ip_add_s<=(others=>'0');
					if timer_enable='1' then
						if init_timer=1 then
							do_tx_dhcp<='1';
							init_timer<=0;
							dhcp_type<=DHCPDISCOVER;
							state_dhcp<=selecting;
						else
							init_timer<=init_timer+1;
						end if;
					end if;

				when selecting =>
					if rcv_dhcp_offer='1' then
						do_tx_dhcp<='1';
						init_timer<=0;
						dhcp_type<=DHCPREQUEST;
						secs<=(others=>'0');
						state_dhcp<=requesting;

					else
						do_tx_dhcp<='0';
						if timer_enable='1' then
							secs<=secs+1;
							if init_timer=2 then
								init_timer<=0;
								state_dhcp<=init;
							else
								init_timer<=init_timer+1;
							end if;
						end if;
					end if;

				when requesting =>
					timer<=(others=>'0');

					if rcv_dhcp_ack='1' then
						server_ip_add_s<=rcv_server_ip_add_s;
						client_ip_add_s<=offered_client_ip_add_s;
						T1<='0' & lease_time_offered(31 downto 1); --50% of lease time
						T2<= lease_time_offered - ("000" & lease_time_offered(31 downto 3));--87.5% of lease time (1 -1/8)*lease_time_offered
						init_timer<=0;
						state_dhcp<=bound;
						secs<=(others=>'0');
					elsif rcv_dhcp_nack='1' then
						init_timer<=0;
						state_dhcp<=init;
					else
						do_tx_dhcp<='0';
						if timer_enable='1' then
							secs<=secs+1;
							if init_timer=10 then
								init_timer<=0;
								state_dhcp<=init;
							else
								init_timer<=init_timer+1;
							end if;
						end if;
					end if;

				when bound =>
					if timer_enable='1' then
						if timer=T1 then
							do_tx_dhcp<='1';
							dhcp_type<=DHCPREQUEST;
							state_dhcp<=renewing;
						else
							do_tx_dhcp<='0';
							timer<=timer+1;
						end if;
					else
						do_tx_dhcp<='0';
					end if;

				when renewing =>
					if rcv_dhcp_ack='1' then
						timer<=(others=>'0');
						state_dhcp<=bound;
						secs<=(others=>'0');
					elsif rcv_dhcp_nack='1' then
						init_timer<=0;
						secs<=(others=>'0');
						state_dhcp<=init;
					elsif timer_enable='1' then
						secs<=secs+1;
						if timer=T2 then
							do_tx_dhcp<='1';
							dhcp_type<=DHCPREQUEST;
							server_ip_add_s<=(others=>'1');
							state_dhcp<=rebinding;
						elsif init_timer=10 then --regular renew trial (every 10s)
							do_tx_dhcp<='1';
							dhcp_type<=DHCPREQUEST;
							init_timer<=0;
						else
							do_tx_dhcp<='0';
							timer<=timer+1;
							init_timer<=init_timer+1;
						end if;
					else
						do_tx_dhcp<='0';
					end if;

				when rebinding =>
					if rcv_dhcp_ack='1' then
						client_ip_add_s<=offered_client_ip_add_s;
						T1<='0' & lease_time_offered(31 downto 1); --50% of lease time
						T2<= lease_time_offered - ("000" & lease_time_offered(31 downto 3));--87.5% of lease time (1 -1/8)*lease_time_offered
						server_ip_add_s<=rcv_server_ip_add_s;
						timer<=(others=>'0');
						secs<=(others=>'0');
						state_dhcp<=bound;
					elsif rcv_dhcp_nack='1' then
						init_timer<=0;
						secs<=(others=>'0');
						state_dhcp<=init;
					elsif timer_enable='1' then
						secs<=secs+1;
						if timer=lease_time_offered then
							init_timer<=0;
							server_ip_add_s<=(others=>'1');
							state_dhcp<=init;
						elsif init_timer=10 then --regular rebind trial (every 10s)
							do_tx_dhcp<='1';
							dhcp_type<=DHCPREQUEST;
							init_timer<=0;
						else
							do_tx_dhcp<='0';
							timer<=timer+1;
							init_timer<=init_timer+1;
						end if;
					else
						do_tx_dhcp<='0';
					end if;
			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- DHCP TX state machine
-------------------------------------------------------------------------------
src_udp_tx(1).hdr.src_port<=std_logic_vector(to_unsigned(68,16));
src_udp_tx(1).hdr.checksum<=(others=>'0');
src_udp_tx(1).hdr.dst_port<=std_logic_vector(to_unsigned(67,16));

process(clk)
begin
	if rising_edge(clk) then
		if reset='1' then
			src_udp_tx_start(1)<='0';
			src_udp_tx(1).data.data_out_valid<='0';
			src_udp_tx(1).data.data_out_last<='0';
			src_udp_tx(1).data.data_out<=(others=>'0');
			src_udp_tx(1).hdr.data_length<=(others=>'0');
			
			state_dhcp_tx<=idle;
		else
			case state_dhcp_tx is
				when idle =>
					tx_count<=0;
					src_udp_tx(1).hdr.dst_ip_addr<=server_ip_add_s;
					dhcp_type2xmit<=dhcp_type;
					if do_tx_dhcp='1' then
						case dhcp_type is
							when DHCPREQUEST =>
								src_udp_tx(1).hdr.data_length<=std_logic_vector(to_unsigned(240+22,16));
							when DHCPDISCOVER =>
								src_udp_tx(1).hdr.data_length<=std_logic_vector(to_unsigned(240+10,16));
						end case;
						state_dhcp_tx<=set_len;
					end if;

				when set_len =>
					src_udp_tx_start(1)<='1';
					src_udp_tx(1).data.data_out_valid<='1';
					src_udp_tx(1).data.data_out<=X"01"; -- OP --request is 01, reply is 02
					tx_count<=tx_count+1;
					state_dhcp_tx<=xmit_bootp;

				when xmit_bootp =>
					if udp_tx_result_dhcp = UDPTX_RESULT_ERR then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						state_dhcp_tx<=IDLE;

					elsif udp_tx_data_out_ready_dhcp='1' then
						if tx_count=239 then
							tx_count<=0;
							if dhcp_type2xmit=dhcpdiscover then
								state_dhcp_tx<=xmit_dhcp_discover;
							else
								state_dhcp_tx<=xmit_dhcp_request;
							end if;
						else
							tx_count<=tx_count+1;
						end if;

						case tx_count is
							when 1  => src_udp_tx(1).data.data_out<=X"01"; -- HTYPE
							when 2  => src_udp_tx(1).data.data_out<=X"06"; -- HLEN
							when 3  => src_udp_tx(1).data.data_out<=X"00"; -- HOPS
							
							when 4  => src_udp_tx(1).data.data_out<=X"39"; -- XID
							when 5  => src_udp_tx(1).data.data_out<=X"03"; -- XID
							when 6  => src_udp_tx(1).data.data_out<=X"F3"; -- XID
							when 7  => src_udp_tx(1).data.data_out<=X"26"; -- XID
							
							when 8  => src_udp_tx(1).data.data_out<=std_logic_vector(secs(15 downto 8)); -- SECS
							when 9  => src_udp_tx(1).data.data_out<=std_logic_vector(secs(7 downto 0)); -- SECS
							when 10  => src_udp_tx(1).data.data_out<=X"00"; -- FLAGS
							when 11  => src_udp_tx(1).data.data_out<=X"00"; -- FLAGS
							
							when 12  => src_udp_tx(1).data.data_out<=client_ip_add_s(31 downto 24); -- current IP
							when 13 => src_udp_tx(1).data.data_out<=client_ip_add_s(23 downto 16); -- current IP
							when 14  => src_udp_tx(1).data.data_out<=client_ip_add_s(15 downto 8); -- current IP
							when 15  => src_udp_tx(1).data.data_out<=client_ip_add_s(7 downto 0); -- current IP

-- 							when 16  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(31 downto 24); -- next IP
-- 							when 17  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(23 downto 16); -- next IP
-- 							when 18  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(15 downto 8); -- next IP
-- 							when 19  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(7 downto 0);  -- next IP
-- 
							when 16  => src_udp_tx(1).data.data_out<=x"00";
							when 17  => src_udp_tx(1).data.data_out<=x"00"; -- next IP
							when 18  => src_udp_tx(1).data.data_out<=x"00"; -- next IP
							when 19  => src_udp_tx(1).data.data_out<=x"00"; -- next IP

-- 							when 20  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(31 downto 24); -- server IP
-- 							when 21  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(23 downto 16); -- server IP
-- 							when 22  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(15 downto 8); -- server IP
-- 							when 23  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(7 downto 0);  -- server IP
-- 
							when 20  => src_udp_tx(1).data.data_out<=X"00"; -- server IP
							when 21  => src_udp_tx(1).data.data_out<=X"00"; -- server IP
							when 22  => src_udp_tx(1).data.data_out<=X"00"; -- server IP
							when 23  => src_udp_tx(1).data.data_out<=X"00";  -- server IP

							when 24  => src_udp_tx(1).data.data_out<=X"00"; -- GIADDR
							when 25  => src_udp_tx(1).data.data_out<=X"00"; -- GIADDR
							when 26  => src_udp_tx(1).data.data_out<=X"00"; -- GIADDR
							when 27  => src_udp_tx(1).data.data_out<=X"00"; -- GIADDR
							
							when 28  => src_udp_tx(1).data.data_out<=our_mac_address(47 downto 40); -- CHADDR
							when 29  => src_udp_tx(1).data.data_out<=our_mac_address(39 downto 32); -- CHADDR
							when 30  => src_udp_tx(1).data.data_out<=our_mac_address(31 downto 24); -- CHADDR
							when 31  => src_udp_tx(1).data.data_out<=our_mac_address(23 downto 16); -- CHADDR
							when 32  => src_udp_tx(1).data.data_out<=our_mac_address(15 downto 8); -- CHADDR
							when 33  => src_udp_tx(1).data.data_out<=our_mac_address(7 downto 0); -- CHADDR
							
							when 236  => src_udp_tx(1).data.data_out<=x"63"; -- MAGIC COOKIE
							when 237  => src_udp_tx(1).data.data_out<=x"82"; -- MAGIC COOKIE
							when 238  => src_udp_tx(1).data.data_out<=x"53"; -- MAGIC COOKIE
							when 239  => src_udp_tx(1).data.data_out<=x"63"; -- MAGIC COOKIE

							when others  => src_udp_tx(1).data.data_out<=X"00"; -- BOOTP
						end case;
					end if;
				
				when xmit_dhcp_discover =>
					if udp_tx_result_dhcp = UDPTX_RESULT_ERR then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						src_udp_tx_start(1)<='0';
						state_dhcp_tx<=IDLE;

					elsif udp_tx_data_out_ready_dhcp='1' then
						if tx_count=9 then
							tx_count<=0;
							src_udp_tx(1).data.data_out_last<='1';
							state_dhcp_tx<=finish;
						else
							tx_count<=tx_count+1;
						end if;

						case tx_count is
							when 0 =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(53,8)); -- option 53 : DHCP type
							when 1 =>src_udp_tx(1).data.data_out<=x"01"; -- length
							when 2 =>src_udp_tx(1).data.data_out<=x"01"; -- dhcpdiscover

							when 3 =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(51,8)); -- option 51 : IP lease time
							when 4 =>src_udp_tx(1).data.data_out<=x"04"; -- length
							when 5 =>src_udp_tx(1).data.data_out<=lease_time_vect(31 downto 24); -- lease time request
							when 6 =>src_udp_tx(1).data.data_out<=lease_time_vect(23 downto 16); -- lease time request
							when 7 =>src_udp_tx(1).data.data_out<=lease_time_vect(15 downto 8); -- lease time request
							when 8 =>src_udp_tx(1).data.data_out<=lease_time_vect(7 downto 0); -- lease time request
							
							when others =>src_udp_tx(1).data.data_out<=x"FF"; -- option 255 : end
						end case;
					end if;

				when xmit_dhcp_request =>
					if udp_tx_result_dhcp = UDPTX_RESULT_ERR then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						src_udp_tx_start(1)<='0';
						state_dhcp_tx<=IDLE;

					elsif udp_tx_data_out_ready_dhcp='1' then
						if tx_count=21 then
							tx_count<=0;
							src_udp_tx(1).data.data_out_last<='1';
							state_dhcp_tx<=finish;
						else
							tx_count<=tx_count+1;
						end if;

						case tx_count is
							when 0 =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(53,8)); -- option 53 : DHCP type
							when 1 =>src_udp_tx(1).data.data_out<=x"01"; -- length
							when 2 =>src_udp_tx(1).data.data_out<=x"03"; -- dhcprequest

							when 3 =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(51,8)); -- option 51 : IP lease time
							when 4 =>src_udp_tx(1).data.data_out<=x"04"; -- length
							when 5 =>src_udp_tx(1).data.data_out<=lease_time_vect(31 downto 24); -- lease time request
							when 6 =>src_udp_tx(1).data.data_out<=lease_time_vect(23 downto 16); -- lease time request
							when 7 =>src_udp_tx(1).data.data_out<=lease_time_vect(15 downto 8); -- lease time request
							when 8 =>src_udp_tx(1).data.data_out<=lease_time_vect(7 downto 0); -- lease time request
							
							when 9   =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(50,8)); -- option 50 : request IP
							when 10  => src_udp_tx(1).data.data_out<=x"04"; -- length
							when 11  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(31 downto 24); -- requested IP
							when 12  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(23 downto 16); -- requested IP
							when 13  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(15 downto 8); -- requested IP
							when 14  => src_udp_tx(1).data.data_out<=offered_client_ip_add_s(7 downto 0);  -- requested IP
							
							when 15   =>src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(54,8)); -- option 54 : server IP
							when 16  => src_udp_tx(1).data.data_out<=x"04"; -- length
							when 17  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(31 downto 24); -- server IP
							when 18  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(23 downto 16); -- server IP
							when 19  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(15 downto 8); -- server IP
							when 20  => src_udp_tx(1).data.data_out<=rcv_server_ip_add_s(7 downto 0);  -- server IP
							
							when others =>src_udp_tx(1).data.data_out<=x"FF"; -- option 255 : end
						end case;
					end if;

				when finish =>
					if udp_tx_data_out_ready_dhcp='1' then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						src_udp_tx_start(1)<='0';
						state_dhcp_tx<=IDLE;
					end if;
			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- DHCP RX state machine
-------------------------------------------------------------------------------
process(clk)
begin
	if rising_edge(clk) then
		if reset='1' then
			rcv_server_ip_add_s<=(others=>'0');
			offered_client_ip_add_s<=(others=>'0');
			rcv_dhcp_offer<='0';
			rcv_dhcp_nack<='0';
			rcv_dhcp_ack<='0';
			state_dhcp_rx<=idle;
		else
			case state_dhcp_rx is
				when idle =>
					option_len<=(others=>'0');
					rx_count<=0;
					dhcp_message_type<=(others=>'1');
					rcv_dhcp_offer<='0';
					rcv_dhcp_nack<='0';
					rcv_dhcp_ack<='0';
					if udp_rxo.hdr.is_valid ='1' and udp_rxo.hdr.dst_port=std_logic_vector(to_unsigned(68,16)) then
						state_dhcp_rx<=rcv_bootp;
					end if;

				when rcv_bootp =>
					if udp_rxo.data.data_in_valid='1' then
						rx_count<=rx_count+1;
						case rx_count is
							when 16 => offered_client_ip_add_s(31 downto 24) <= udp_rxo.data.data_in;
							when 17 => offered_client_ip_add_s(23 downto 16) <= udp_rxo.data.data_in;
							when 18 => offered_client_ip_add_s(15 downto 8) <= udp_rxo.data.data_in;
							when 19 => offered_client_ip_add_s(7 downto 0) <= udp_rxo.data.data_in;
							
-- 							when 20 => rcv_server_ip_add_s(31 downto 24) <= udp_rxo.data.data_in;
-- 							when 21 => rcv_server_ip_add_s(23 downto 16) <= udp_rxo.data.data_in;
-- 							when 22 => rcv_server_ip_add_s(15 downto 8)  <= udp_rxo.data.data_in;
-- 							when 23 => rcv_server_ip_add_s(7 downto 0)   <= udp_rxo.data.data_in;

							when 28 => 
								if (udp_rxo.data.data_in/=our_mac_address(47 downto 40)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 29 => 
								if (udp_rxo.data.data_in/=our_mac_address(39 downto 32)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 30 => 
								if (udp_rxo.data.data_in/=our_mac_address(31 downto 24)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 31 => 
								if (udp_rxo.data.data_in/=our_mac_address(23 downto 16)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 32 => 
								if (udp_rxo.data.data_in/=our_mac_address(15 downto 8)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 33 => 
								if (udp_rxo.data.data_in/=our_mac_address(7 downto 0)) then
									state_dhcp_rx<=wait_end;
								end if;

							when 236  => 
								if (udp_rxo.data.data_in/=x"63") then -- MAGIC COOKIE
									state_dhcp_rx<=wait_end;
								end if;
							when 237  => 
								if (udp_rxo.data.data_in/=x"82") then -- MAGIC COOKIE
									state_dhcp_rx<=wait_end;
								end if;
							when 238  => 
								if (udp_rxo.data.data_in/=x"53") then -- MAGIC COOKIE
									state_dhcp_rx<=wait_end;
								end if;
							when 239  => 
								if (udp_rxo.data.data_in/=x"63") then -- MAGIC COOKIE
									state_dhcp_rx<=wait_end;
								else
									state_dhcp_rx<=rcv_option_type;
								end if;
							when others => null;
						end case;
					end if;
				
				when rcv_option_type =>
					if udp_rxo.data.data_in_valid='1' then
						option<=udp_rxo.data.data_in;
						case udp_rxo.data.data_in is
							when x"00" => null; -- gap filler, no length
							when x"FF" => -- end, no length
								if udp_rxo.data.data_in_last='1' then
									state_dhcp_rx<=finish;
								else
									state_dhcp_rx<=wait_end;
								end if;

							when others =>
								if udp_rxo.data.data_in_last='1' then
									state_dhcp_rx<=idle;
								else
									state_dhcp_rx<=rcv_option_len;
								end if;
						end case;
					end if;

				when rcv_option_len =>
					if udp_rxo.data.data_in_valid='1' then
						option_len<=unsigned(udp_rxo.data.data_in);
						if udp_rxo.data.data_in_last='1' then
							state_dhcp_rx<=idle;
						else
							state_dhcp_rx<=rcv_option_data;
						end if;
					end if;

				when rcv_option_data =>
					if udp_rxo.data.data_in_valid='1' then
						if udp_rxo.data.data_in_last='1' then
								state_dhcp_rx<=idle;
						else
							if option_len=1 then
								state_dhcp_rx<=rcv_option_type;
							else
								option_len<=option_len-1;
							end if;

							if option=std_logic_vector(to_unsigned(51,8)) then --lease time
								lease_time_offered<=lease_time_offered(23 downto 0) & unsigned(udp_rxo.data.data_in);

							elsif option=std_logic_vector(to_unsigned(53,8)) then --DHCP type
								dhcp_message_type<=udp_rxo.data.data_in;

							elsif option=std_logic_vector(to_unsigned(54,8)) then --DCHP server IP
								rcv_server_ip_add_s<=rcv_server_ip_add_s(23 downto 0) & udp_rxo.data.data_in;
							end if;
						end if;
					end if;
				
				when wait_end =>
					if udp_rxo.data.data_in_valid='1' and  udp_rxo.data.data_in_last='1' then
						state_dhcp_rx<=finish;
					end if;
				
				when finish =>
					case dhcp_message_type is
						when x"01" => null;
						when x"02" =>
							rcv_dhcp_offer<='1';
						when x"03" => null;
						when x"04" => null;
						when x"05" =>
							rcv_dhcp_ack<='1';
						when x"06" =>
							rcv_dhcp_nack<='1';
						when x"07" => null;
						when x"08" => null;
						when others => null;
					end case;
					state_dhcp_rx<=idle;
			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- UDP arbitrer
-------------------------------------------------------------------------------
src_udp_tx_start(0)<=udp_tx_start_prior;
src_udp_tx(0)<=udp_txi_prior;

process(clk)
begin
	if rising_edge(clk) then
		if reset = '1' then
			sel<=0;
			state_arb_udp<=idle;
		else
			case state_arb_udp is
				when idle =>
					if src_udp_tx_start(sel)='1' then
						state_arb_udp<=busy;
					elsif sel=nb_src-1 then
						sel<=0;
					else
						sel<=sel+1;
					end if;

				when busy =>
					if src_udp_tx(sel).data.data_out_last='1' then
						state_arb_udp<=idle;
					end if;
			end case;
		end if;
	end if;
end process;

udp_tx_result_prior<=udp_tx_result;
udp_tx_result_dhcp<=udp_tx_result;

udp_tx_data_out_ready_prior<=udp_tx_data_out_ready when sel=0 else '0';
udp_tx_data_out_ready_dhcp <=udp_tx_data_out_ready when sel=1 else '0';

udp_tx_start<=src_udp_tx_start(sel);
udp_txi <= src_udp_tx(sel);

end RTL;

