library std;
use std.textio.all;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_textio.all;

use work.pack_ipv4_types.all;
use work.ipbus.all;

entity tb_top_ipbus_lpsc_extphy_with_arp is
generic (
	PC_MAC_NUMBER_g	: std_logic_vector(47 downto 0):=x"0000_1cb5_ca61";
	PC_IP_ADDRESS_g   : std_logic_vector(31 downto 0):=X"c0a864" & X"09"; -- 192.168.100.09
	PC_IPBUSPORT_g  : std_logic_vector(15 downto 0):=x"C069";

	ELEC_MAC_NUMBER : std_logic_vector(47 downto 0):=x"0022_8f03_0001"; --CENTRAL 
	ELEC_IPBUSPORT  : std_logic_vector(15 downto 0):=x"C351";
	ELEC_IP_ADDRESS   : std_logic_vector(31 downto 0):=X"c0a864" & X"11" -- 192.168.100.17
	);
end tb_top_ipbus_lpsc_extphy_with_arp;

architecture simu of tb_top_ipbus_lpsc_extphy_with_arp is
component top_ipbus_lpsc_extphy
generic (
	IN_SIMULATION     : boolean :=false;
	SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
	CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
	ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
	ARP_MAX_PKT_TMO	: integer := 5;									-- # wrong nwk pkts received before set error
	MAX_ARP_ENTRIES 	: integer := 255									-- max entries in the ARP store
	);
port(
	sysclk_p: in STD_LOGIC;
   	sysclk_n: in STD_LOGIC;

	leds: out STD_LOGIC_VECTOR(3 downto 0);

	gmii_gtx_clk : out STD_LOGIC;
	gmii_tx_en : out STD_LOGIC;
	gmii_tx_er : out STD_LOGIC;
	gmii_txd : out STD_LOGIC_VECTOR(7 downto 0);

	gmii_rx_clk: in STD_LOGIC;
	gmii_rx_dv: in STD_LOGIC;
	gmii_rx_er: in STD_LOGIC;
	gmii_rxd : in STD_LOGIC_VECTOR(7 downto 0);

	phy_rstb : out STD_LOGIC;
	dip_switch: in std_logic_vector(3 downto 0)
	);
end component;

component UDP_Complete
generic (
	IN_SIMULATION     : boolean :=false;
	SIMUL_DST_MAC_NUMBER  : std_logic_vector(47 downto 0):=x"AAAA_BBBB_CCCC";
	CLOCK_FREQ			: integer := 125000000;							-- freq of data_in_clk -- needed to timout cntr
	ARP_TIMEOUT			: integer := 60;									-- ARP response timeout (s)
	ARP_MAX_PKT_TMO	: integer := 5;									-- # wrong nwk pkts received before set error
	MAX_ARP_ENTRIES 	: integer := 255									-- max entries in the ARP store
	);
Port (
	-- UDP TX signals
	udp_tx_start			: in std_logic;							-- indicates req to tx UDP
	udp_txi					: in udp_tx_type;							-- UDP tx cxns
	udp_tx_result			: out std_logic_vector (1 downto 0);-- tx status (changes during transmission)
	udp_tx_data_out_ready: out std_logic;							-- indicates udp_tx is ready to take data

	-- UDP RX signals
	udp_rx_start			: out std_logic;							-- indicates receipt of udp header
	udp_rxo					: out udp_rx_type;

	-- IP RX signals
	ip_rx_hdr				: out ipv4_rx_header_type;

	-- system signals
	clk125					: in  STD_LOGIC;
	reset 					: in  STD_LOGIC;
	our_ip_address 		: in STD_LOGIC_VECTOR (31 downto 0);
	our_mac_address 		: in std_logic_vector (47 downto 0);
	control					: in udp_control_type;

	-- status signals
	arp_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- count of arp pkts received
	ip_pkt_count			: out STD_LOGIC_VECTOR(7 downto 0);			-- number of IP pkts received for us
	
	-- gmii TX interface
	-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
	gmii_tx_en : out STD_LOGIC;
	gmii_tx_er : out STD_LOGIC;
	gmii_txd : out STD_LOGIC_VECTOR(7 downto 0);

	-- gmii RX interface
	gmii_rx_clk: in STD_LOGIC;
   	gmii_rx_dv: in STD_LOGIC;
	gmii_rx_er: in STD_LOGIC;
	gmii_rxd : in STD_LOGIC_VECTOR(7 downto 0)
	);
end component;

component mac_tx_engine
port (
	rstn : in std_logic;
	
	-- to user interface
	clk : in std_logic;
	tx_tdata : in std_logic_vector(7 downto 0); --!frame data to be Xfrerd
	tx_tvalid : in std_logic; --! TX data valid
	tx_tlast : in std_logic; --! final Xfer of frame
	tx_tuser : in std_logic; --! Signal tx error
	tx_tready : out std_logic; --! Transmitter ready, qualifies when data is taken by MAC
	tx_statistics_vector : out std_logic_vector(31 downto 0); 
	tx_statistics_valid : out std_logic;
	tx_mac_config_vector : in std_logic_vector(79 downto 0);

	-- to gmii
	gmii_txd : out std_logic_vector(7 downto 0);
	gmii_tx_en : out std_logic;
	gmii_tx_er : out std_logic
	);
end component;

signal sysclk_p: std_logic;
signal sysclk_n: std_logic;

signal clk125: std_logic;
signal rst_125: std_logic;
signal rstn: std_logic;

signal gmii_tx_clk :std_logic;
signal gmii_tx_en :std_logic;
signal gmii_tx_er :std_logic;
signal gmii_txd :std_logic_vector(7 downto 0);

signal gmii_tx_clk_dhcp :std_logic;
signal gmii_tx_en_dhcp :std_logic;
signal gmii_tx_er_dhcp :std_logic;
signal gmii_txd_dhcp :std_logic_vector(7 downto 0);

signal gmii_tx_clk_ipbus :std_logic;
signal gmii_tx_en_ipbus :std_logic;
signal gmii_tx_er_ipbus :std_logic;
signal gmii_txd_ipbus :std_logic_vector(7 downto 0);

signal tx_tdata : std_logic_vector(7 downto 0);
signal tx_tvalid : std_logic;
signal tx_tlast : std_logic;
signal tx_tuser : std_logic;
signal tx_tready : std_logic;
signal tx_mac_config_vector : std_logic_vector(79 downto 0):=(others=>'0');

signal gmii_rx_clk:std_logic;
signal gmii_rx_dv:std_logic;
signal gmii_rx_er:std_logic;
signal gmii_rxd :std_logic_vector(7 downto 0);
signal is_arp_switch : boolean;
signal send_ack : boolean;
signal send_arp_8 : boolean;
signal arp8_pack_sent: boolean;
signal send_arp_9 : boolean;
signal arp9_pack_sent: boolean;

signal PC_MAC_NUMBER	: std_logic_vector(47 downto 0):=x"0000_1cb5_ca61";
signal PC_IP_ADDRESS   : std_logic_vector(31 downto 0):=X"c0a864" & X"09"; -- 192.168.100.09
signal PC_IPBUSPORT  : std_logic_vector(15 downto 0):=x"C069";

signal PC_MAC_NUMBER_alt : std_logic_vector(47 downto 0):=x"0000_1cb5_ca60";
signal PC_IP_ADDRESS_alt : std_logic_vector(31 downto 0):=X"c0a864" & X"20"; -- 192.168.100.64
signal PC_IPBUSPORT_alt  : std_logic_vector(15 downto 0):=x"C060";
--UDP/IPBUS interconnection
signal udp_tx_start			:std_logic;							-- indicates req to tx UDP
-- signal udp_tx_start_cern			:std_logic;							-- indicates req to tx UDP
signal udp_txi	: udp_tx_type;
signal udp_tx_result		: std_logic_vector (1 downto 0);-- tx status (changes during transmission)
-- signal udp_tx_result_cern		: std_logic_vector (1 downto 0);-- tx status (changes during transmission)
signal udp_tx_data_out_ready: std_logic;							-- indicates udp_tx is ready to take data
-- signal udp_tx_data_out_ready_cern: std_logic;							-- indicates udp_tx is ready to take data

signal udp_rx_start	: std_logic;							-- indicates receipt of udp header
signal udp_rxo		: udp_rx_type;
-- signal udp_rx_start_cern	: std_logic;							-- indicates receipt of udp header
-- signal udp_rxo_cern		: udp_rx_type;

signal control			: udp_control_type;

-- testbench signals
signal rxbuf_sel : natural;
signal rxdat_sel : natural;

signal txbuf_sel : natural;
signal txdat_sel : natural;
------------------------------------------------------------------------------
-- types to support transaction buffer data
------------------------------------------------------------------------------
type data_typ is array (natural range <>) of bit_vector(31 downto 0);
type frame_typ_ary is array (natural range <>) of data_typ(0 to 40);

signal tx_buf : frame_typ_ary(0 to 18);

constant transaction_data : frame_typ_ary := (
-------------
-- Frame 0 
-------------
0          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0001_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"A000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 1 --same as before, good number
-------------
1          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"BC00_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 2 --same as before, bad packet number
-------------
2          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 3 --same as before, 0 packet number
-------------
3          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0000_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 4 --same as before, good packet number
-------------
4          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0003_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1000",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_01_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 5 --same as before, good packet number, throughput measurement
-------------
5          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0004_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 6 --same as before, good packet number, throughput measurement
-------------
6          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0005_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),


7          => (
	others        => x"FFFF_FFFF") 
);
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
constant stat_resend_data : frame_typ_ary := ( -- expect 6 packets
-------------
-- Frame 0 
-------------
0          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0001_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"A000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 1 --same as before, good number
-------------
1          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"BC00_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 2 --same as before, bad packet number
-------------
2          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 3 --same as before, 0 packet number
-------------
3          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0000_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 4 --same as before, good packet number
-------------
4          => (
	0        => to_bitvector(conv_std_logic_vector(24,32)),    --packet_size
	1        => X"20_0003_F0",   --first ipbus packet header
	
	2        => X"2_000_01_1_F", --first transaction (single write)
	3        => X"0000_1000",    --first transaction base address
	4        => X"ABBA_B000",    --first transaction 1 data
	
	5        => X"2_001_03_1_F", --2nd transaction (triple write)
	6        => X"0000_1001",    --2nd transaction base address
	7        => X"ABCA_C001",    --2nd transaction data 1
	8        => X"ABCA_C002",    --2nd transaction data 2
	9        => X"ABCA_C003",    --2nd transaction data 3

	10       => X"2_002_03_0_F", --3rd transaction (triple read)
	11       => X"0000_1000",    --3rd transaction base address
	
	12       => X"2_003_04_2_F", --4th transaction (quad ni_read)
	13       => X"0000_1000",    --4th transaction base address
	
	14       => X"2_004_02_3_F", --transaction (double ni_write)
	15       => X"0000_1010",    --transaction base address
	16       => X"A000_D001",    --transaction data 1
	17       => X"A000_D002",    --transaction data 2

	18       => X"2_005_01_4_F", --transaction (RMW bit)
	19       => X"0000_1010",    --transaction base address
	20       => X"0000_F000",    --AND term
	21       => X"000A_0000",    --OR term

	22       => X"2_006_01_5_F", --transaction (RMW sum)
	23       => X"0000_1010",    --transaction base address
	24       => X"0000_00B0",    --ADDEND
	others   => X"0000_0000"),

-------------
-- Frame 5 --status packet
-------------
5          => (
	0        => to_bitvector(conv_std_logic_vector(17,32)),    --packet_size
	1        => X"20_0000_F1",   --first ipbus packet header
	
	others   => X"0000_0000"),

-------------
-- Frame 6 --se-send ipbus packet 0001
-------------
6          => (
	0        => to_bitvector(conv_std_logic_vector(1,32)),    --packet_size
	1        => X"20_0002_F2",   --first ipbus packet header
	others   => X"0000_0000"),


7          => (
	others        => x"FFFF_FFFF") 
);
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
constant throughput_buf : frame_typ_ary := (
-------------
-- Frame 0 -- special packet ID of 0 -- init mem content of 2 cells
-------------
0          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0000_F0",   --first ipbus packet header
	
	2        => X"2_000_02_1_F", --transaction (triple write)
	3        => X"0000_1000",    --transaction base address
	4        => X"BEEF_BABE",    --transaction data 1
	5        => X"ABBA_ABBA",    --transaction data 2
	
	others   => X"0000_0000"),


-------------
-- Frame 1 --same as before, good packet number, throughput measurement
-------------
1          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0001_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1001",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 2 --same as before, good packet number, throughput measurement
-------------
2          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0002_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 3 --same as before, good packet number, throughput measurement
-------------
3          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0003_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 4 --same as before, good packet number, throughput measurement
-------------
4          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0004_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 5 --same as before, good packet number, throughput measurement
-------------
5          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0005_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 6 --same as before, good packet number, throughput measurement
-------------
6          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0006_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 7 --same as before, good packet number, throughput measurement
-------------
7          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0007_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	others   => X"0000_0000"),
	
-------------
-- Frame 8 --same as before, good packet number, throughput measurement
-------------
8          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0008_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 9 --same as before, good packet number, throughput measurement
-------------
9          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0009_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 10 --same as before, good packet number, throughput measurement
-------------
10          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000A_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
-- 	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	4        => X"2_001_5E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 11 --same as before, good packet number, throughput measurement
-------------
11          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000B_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 12 --same as before, good packet number, throughput measurement
-------------
12          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000C_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 13 --same as before, good packet number, throughput measurement
-------------
13          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000D_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 14 --same as before, good packet number, throughput measurement
-------------
14          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000E_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 15 --same as before, good packet number, throughput measurement
-------------
15          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_000F_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 16 --same as before, good packet number, throughput measurement
-------------
16          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0010_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),

-------------
-- Frame 17 --same as before, good packet number, throughput measurement
-------------
17          => (
	0        => to_bitvector(conv_std_logic_vector(5,32)),    --packet_size
	1        => X"20_0011_F0",   --first ipbus packet header
	
	2        => X"2_000_FF_2_F", --first transaction (255 NI read)
	3        => X"0000_1000",    --first transaction base address
	
	4        => X"2_001_6E_2_F", --first transaction (110 NI read)
	5        => X"0000_1000",    --first transaction base address
	
	others   => X"0000_0000"),
-------------
-- Frame 18 --status packet
-------------
18          => (
	0        => to_bitvector(conv_std_logic_vector(17,32)),    --packet_size
	1        => X"20_0000_F1",   --status packet packet header
	
	others   => X"0000_0000"),

-------------
-- Frame 19 --se-send ipbus packet 0000A
-------------
19          => (
	0        => to_bitvector(conv_std_logic_vector(1,32)),    --packet_size
	1        => X"20_000A_F2",   --packet to resend
	others   => X"0000_0000")

);

procedure send_buffer(
	variable current_buf	: natural; 
	variable is_LE			: in boolean;
	constant data_buf		: in frame_typ_ary;

	signal clk125					: in std_logic;
	signal udp_tx_data_out_ready	: in std_logic;
	signal udp_tx_start				: out std_logic;
	signal udp_txi					: out udp_tx_type
) is
variable txsize : natural;
variable byte2use : natural;
begin
	txsize:=conv_integer(to_stdlogicvector(data_buf(current_buf)(0)));
	udp_tx_start<='0';
	
	udp_txi.hdr.dst_port<=ELEC_IPBUSPORT;
	udp_txi.hdr.src_port<=PC_IPBUSPORT;
	udp_txi.hdr.data_length<=conv_std_logic_vector(txsize*4,16);
	udp_txi.hdr.dst_ip_addr<=ELEC_IP_ADDRESS;
	udp_txi.hdr.checksum<=(others=>'0');

	wait until rising_edge(clk125);
	udp_tx_start<='1';
	udp_txi.data.data_out_valid<='1';

	for i in 1 to txsize loop
		for byte in 0 to 3 loop
			if is_LE then
				byte2use:=byte;
			else
				byte2use:=3-byte;
			end if;
			udp_txi.data.data_out<=to_stdlogicvector(data_buf(current_buf)(i)(8*(byte2use+1)-1 downto 8*byte2use));

			if i=txsize and byte=3 then
				udp_txi.data.data_out_last<='1';
				udp_tx_start<='0';
			end if;

			while true loop
				wait until rising_edge(clk125);
				if udp_tx_data_out_ready='1' then
					exit;
				end if;
			end loop;
		end loop;
	end loop;

	udp_txi.data.data_out_last<='0';
	udp_txi.data.data_out_valid<='0';
	wait until rising_edge(clk125);

end send_buffer;


begin

p_gtx_clk : process
	begin
	sysclk_p <= '0';
	sysclk_n <= '1';
	wait for 2.5 ns;

	sysclk_p <= '1';
	sysclk_n <= '0';
	wait for 2.5 ns;
end process p_gtx_clk;

process
	begin
	clk125 <= '0';
	wait for 4 ns;

	clk125 <= '1';
	wait for 4 ns;
end process;
-------------------------------------------------------------------------------
-- DUT
-------------------------------------------------------------------------------
lbl_DUT : top_ipbus_lpsc_extphy generic map(
	IN_SIMULATION=>true,
	SIMUL_DST_MAC_NUMBER=>PC_MAC_NUMBER_g,
	CLOCK_FREQ=>125000000,
	ARP_TIMEOUT=>60,
	ARP_MAX_PKT_TMO=>5,
	MAX_ARP_ENTRIES=>255
	)
port map(
	sysclk_p=>sysclk_p,
   	sysclk_n=>sysclk_n,

	leds=>open,

	gmii_gtx_clk=>gmii_tx_clk,
   	gmii_tx_en=>gmii_tx_en,
	gmii_tx_er=>gmii_tx_er,
	gmii_txd=>gmii_txd,

	gmii_rx_clk=>gmii_rx_clk,
   	gmii_rx_dv=>gmii_rx_dv,
	gmii_rx_er=>gmii_rx_er,
	gmii_rxd=>gmii_rxd,

	phy_rstb=>open,
	dip_switch=>"0011"
	);
gmii_rxd<=gmii_txd_dhcp when is_arp_switch else gmii_txd_ipbus;
gmii_rx_dv<=gmii_tx_en_dhcp when is_arp_switch else gmii_tx_en_ipbus;
gmii_rx_er<=gmii_tx_er_dhcp when is_arp_switch else gmii_tx_er_ipbus;
-------------------------------------------------------------------------------
-- UDP master side (LPSC)
-------------------------------------------------------------------------------
lbl_UDP_tb : UDP_Complete generic map(
		IN_SIMULATION=>true,
		SIMUL_DST_MAC_NUMBER=>ELEC_MAC_NUMBER,
		CLOCK_FREQ=>125000000,
		ARP_TIMEOUT=>60,
		ARP_MAX_PKT_TMO=>5,
		MAX_ARP_ENTRIES=>255
		)
	port map(
		-- UDP TX signals
		udp_tx_start=>udp_tx_start,
		udp_txi=>udp_txi,
		udp_tx_result=>udp_tx_result,
		udp_tx_data_out_ready=>udp_tx_data_out_ready,
		
		-- UDP RX signals
		udp_rx_start=>udp_rx_start,
		udp_rxo=>udp_rxo,
		
		-- IP RX signals
		ip_rx_hdr=>open,

		-- system signals
		clk125=>clk125,
		reset=>rst_125,
		our_ip_address=>PC_IP_ADDRESS,
		our_mac_address=>PC_MAC_NUMBER,
		control=>control,

		-- status signals
		arp_pkt_count=>open,
		ip_pkt_count=>open,
		
		-- gmii TX interface
		-- 	gmii_gtx_clk : out STD_LOGIC; -- must be generated with appropriate system
		gmii_txd=>gmii_txd_ipbus,
		gmii_tx_en=>gmii_tx_en_ipbus,
		gmii_tx_er=>gmii_tx_er_ipbus,

		-- gmii RX interface
		gmii_rx_clk=>gmii_tx_clk,
		gmii_rxd =>gmii_txd,
		gmii_rx_dv=>gmii_tx_en,
		gmii_rx_er=>gmii_tx_er
		);

gmii_rx_clk<=clk125; --for simulation, it is fine 

control.ip_controls.arp_controls.clear_cache<='0';
--------------------------------------------------------------------------------
--  dhcp server
--------------------------------------------------------------------------------
dlink_switch : process
variable l:         line;
variable c:         character;
variable is_string: boolean;

file file_pointer : text;
variable file_line : line;
variable temp_line : line;
variable Message : line;
variable file_name : string(1 to 20);
variable data_int : std_logic_vector(7 downto 0);
begin
	tx_tdata<=(others=>'0');
	tx_tvalid<='0';
	tx_tlast<='0';
	arp8_pack_sent<=false;
	arp9_pack_sent<=false;
	wait until rising_edge(gmii_rx_clk) and send_arp_8;
	
	file_open(file_pointer,"./ipbus_lpsc/ipbus/arp8_reply_sim.txt",read_mode);
-- 	file_open(file_pointer,"./ipbus_lpsc/ipbus/resend_request_sim.txt",read_mode);
	while not(endfile(file_pointer)) loop
		readline(file_pointer,file_line);
		hread(file_line,data_int);
		tx_tvalid<='1';
		if endfile(file_pointer) then
			tx_tlast<='1';
		end if;

		tx_tdata<=data_int;
		while (true) loop
			wait until rising_edge(gmii_rx_clk);
			if tx_tready='1' then
				exit;
			end if;
		end loop;
	end loop;
	file_close(file_pointer);

	tx_tvalid<='0';
	tx_tlast<='0';
	wait until rising_edge(gmii_rx_clk) and gmii_tx_en_dhcp='0';
	arp8_pack_sent<=true;

	---------------------
	---------------------
	wait until rising_edge(gmii_rx_clk) and send_arp_9;
	
	file_open(file_pointer,"./ipbus_lpsc/ipbus/arp9_reply_sim.txt",read_mode);
-- 	file_open(file_pointer,"./ipbus_lpsc/ipbus/resend_request_sim.txt",read_mode);
	while not(endfile(file_pointer)) loop
		readline(file_pointer,file_line);
		hread(file_line,data_int);
		tx_tvalid<='1';
		if endfile(file_pointer) then
			tx_tlast<='1';
		end if;

		tx_tdata<=data_int;
		while (true) loop
			wait until rising_edge(gmii_rx_clk);
			if tx_tready='1' then
				exit;
			end if;
		end loop;
	end loop;
	file_close(file_pointer);

	tx_tvalid<='0';
	tx_tlast<='0';
	wait until rising_edge(gmii_rx_clk) and gmii_tx_en_dhcp='0';
	arp9_pack_sent<=true;

	wait for 1 us;
	wait until rising_edge(gmii_rx_clk);
end process dlink_switch;

lbl_dhcp_mac : mac_tx_engine port map(
	rstn=>rstn,
	
	-- to user interface
	clk=>gmii_rx_clk,
	tx_tdata=>tx_tdata,
	tx_tvalid=>tx_tvalid,
	tx_tlast=>tx_tlast,
	tx_tuser=>'0',
	tx_tready=>tx_tready,

	tx_statistics_vector=>open,
	tx_statistics_valid=>open,
	tx_mac_config_vector=>tx_mac_config_vector,

	-- to gmii
	gmii_txd=>gmii_txd_dhcp,
	gmii_tx_en=>gmii_tx_en_dhcp,
	gmii_tx_er=>gmii_tx_er_dhcp
	);
rstn<=not(rst_125);
-------------------------------------------------------------------------------
-- main process
-------------------------------------------------------------------------------
main : process
variable idx : natural;
variable LE_endianess : boolean;
begin
	udp_tx_start<='0';
	udp_txi.data.data_out_valid<='0';
	udp_txi.data.data_out_last<='0';
	udp_txi.data.data_out<=(others=>'0');
	
	PC_MAC_NUMBER<=PC_MAC_NUMBER_g;
	PC_IP_ADDRESS<=PC_IP_ADDRESS_g;
	PC_IPBUSPORT<=PC_IPBUSPORT_g;
	wait for 1 ns;

	is_arp_switch<=false;
	send_arp_8<=false;

	rst_125<='1';
	wait for 100 ns;

	rst_125<='0';
	wait for 200 ns;
	LE_endianess:=true;

	PC_MAC_NUMBER<=PC_MAC_NUMBER_g;
	PC_IP_ADDRESS<=PC_IP_ADDRESS_g;
	PC_IPBUSPORT<=PC_IPBUSPORT_g;
	wait for 1 ns;

	send_buffer(current_buf=>18,data_buf=>throughput_buf,
			is_LE=>LE_endianess,
			clk125=>clk125,udp_tx_data_out_ready=>udp_tx_data_out_ready,
			udp_tx_start=>udp_tx_start,udp_txi=>udp_txi);

			
	wait for 2 us;



	PC_MAC_NUMBER<=PC_MAC_NUMBER_alt;
	PC_IP_ADDRESS<=PC_IP_ADDRESS_alt;
	PC_IPBUSPORT<=PC_IPBUSPORT_alt;
	wait for 1 ns;
	send_buffer(current_buf=>18,data_buf=>throughput_buf,
			is_LE=>LE_endianess,
			clk125=>clk125,udp_tx_data_out_ready=>udp_tx_data_out_ready,
			udp_tx_start=>udp_tx_start,udp_txi=>udp_txi);

	wait for 2 us;


-- 	is_arp_switch<=true;
-- 	send_arp_8<=true;
-- 
-- 	wait until arp8_pack_sent;
-- 	send_arp_8<=false;
-- 	is_arp_switch<=false;
-- 	wait for 16 ns;

	wait for 1 us;

-- 	 wait for 50 us;
	report "fin de simulation" severity ERROR;
end process main;
end simu;
