-- ntp_udp_client
--
-- manage the communication with the UDP layer.
-- Use local time and provides T'1,T'2, T1 and T2
--
-- poll interval set to 128s (exponent of 7)
-- timeout value is equal to the poll interval
--
-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity ntp_udp_client is 
generic(
	IN_SIMULATION :  boolean:=false; -- when true, time_enable period is  accelerated by 200000 (5 us instead of 1s)
	REFCLK_FREQ :  natural:=10E6; -- default to 10MHz
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);
	enable_ntp : in std_logic;

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk)
	refclk: in std_logic;
	rst_refclk: in std_logic; -- referenced to refclk
	new_time_avail: in std_logic; -- signal to use for clock domain synchronisation
	local_time : in std_logic_vector(63 downto 0);
	reference_time : in std_logic_vector(63 downto 0); -- last time when system was updated

	ntp_frame_rcvd: out std_logic;
	ntp_timeout: out std_logic; --a time out if 4 successive poll failed
	T_1 : out std_logic_vector(63 downto 0); -- rcvd OT field
	T_2 : out std_logic_vector(63 downto 0); -- local time at datagram rcv time
	T_prime_1 : out std_logic_vector(63 downto 0); -- rcvd RT field
	T_prime_2 : out std_logic_vector(63 downto 0) -- rcvd TT field
	);
end ntp_udp_client;

architecture RTL of ntp_udp_client is
component ntp_udp_rx
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	local_time : in std_logic_vector(63 downto 0); -- local time
	local_time_at_rcv : out std_logic_vector(63 downto 0); -- local time at datagram rcv time

	rcv_ntp_ack: out std_logic;
	origin_time : out std_logic_vector(63 downto 0); -- rcvd OT field
	receive_time : out std_logic_vector(63 downto 0); -- rcvd RT field
	Transmit_time : out std_logic_vector(63 downto 0) -- rcvd TT field
	);
end component ntp_udp_rx;

component ntp_udp_tx
generic(
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);

	do_request_ntp : in std_logic;
	origin_time : in std_logic_vector(63 downto 0);
	receive_time : in std_logic_vector(63 downto 0);
	transmit_time : in std_logic_vector(63 downto 0); -- local time (hence transmit time)
	reference_time : in std_logic_vector(63 downto 0); -- last time when system was updated

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic
	);
end component ntp_udp_tx;

-- freq_scaler
signal freq_scaler :natural range 0 to REFCLK_FREQ-1;
signal timer_enable : std_logic;

-- NTP main manager
signal we_have_ip : std_logic;
signal we_have_ntp_serv : std_logic;

type def_state_ntp_main is (idle,ntp_polling);
signal state_ntp_main : def_state_ntp_main;
signal do_ntp : std_logic;
signal do_request_ntp : std_logic;
signal ntp_frame_rcvd_s: std_logic;
signal no_ntp_timeout : std_logic;
signal ntp_tmout_cnt : natural range 0 to 3;
signal ntp_polling_cnt : natural range 0 to 2**poll_interval_exp-1;

signal new_time_avail_1ck: std_logic;
signal new_time_avail_2ck: std_logic;
signal local_time_sync : std_logic_vector(63 downto 0);
signal reference_time_sync : std_logic_vector(63 downto 0);
signal do_ntp_1ck : std_logic;
signal do_ntp_2ck : std_logic;
constant stretch_dur : natural:= (4*125E6)/REFCLK_FREQ;
signal  stretch_cnt : natural range 0 to stretch_dur-1;
type def_state_stretch is (idle,stretching);
signal state_stretch : def_state_stretch;
signal rcv_ntp_ack: std_logic;
signal rcv_ntp_ack_stretch : std_logic;
signal rcv_ntp_ack_stretch_1ck : std_logic;
signal rcv_ntp_ack_stretch_2ck : std_logic;

begin
-------------------------------------------------------------------------------
-- FREQ scaler
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if rst_refclk = '1' then
			if IN_SIMULATION then 
				freq_scaler <= REFCLK_FREQ/200000-1;
			else
				freq_scaler <= REFCLK_FREQ-1;
			end if;
			timer_enable<='0';
		else
			-- freq scaling and 1-sec timer
			if freq_scaler = 0 then
				if IN_SIMULATION then 
					freq_scaler <= REFCLK_FREQ/200000-1;
				else
					freq_scaler <= REFCLK_FREQ-1;
				end if;
				timer_enable<='1';
			else
				freq_scaler <= freq_scaler - 1;
				timer_enable<='0';
			end if;
		end if;
	end if;
end process;

process(mac_clk)
begin
	if rising_edge(mac_clk) then
		if rst_macclk = '1' then
			rcv_ntp_ack_stretch<='0';
			state_stretch<=idle;
		else
			case state_stretch is
				when idle =>
					stretch_cnt<=stretch_dur-1;
					if rcv_ntp_ack='1' then
						rcv_ntp_ack_stretch<='1';
						state_stretch<=stretching;
					else
						rcv_ntp_ack_stretch<='0';
					end if;
				when stretching =>
					if stretch_cnt=0 then
						rcv_ntp_ack_stretch<='0';
						state_stretch<=idle;
					else
						stretch_cnt<=stretch_cnt-1;
					end if;
			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- NTP main state machine (refclk)
-------------------------------------------------------------------------------
process(refclk)
begin
	if rising_edge(refclk) then
		if our_ip_address=x"00_00_00_00" then
			we_have_ip<='0';
		else
			we_have_ip<='1';
		end if;

		if ntp_server_add=x"00_00_00_00" then
			we_have_ntp_serv<='0';
		else
			we_have_ntp_serv<='1';
		end if;

		if rst_refclk = '1' or enable_ntp='0' then
			ntp_timeout<='0';
			do_ntp<='0';
			state_ntp_main<=idle;
		else
			case state_ntp_main is 
				when idle =>
					ntp_polling_cnt<=2**poll_interval_exp-1;
					no_ntp_timeout<='0';
					if timer_enable='1' then
						if we_have_ip='1' and we_have_ntp_serv='1' then
							do_ntp<='1';
							state_ntp_main<=ntp_polling;
						end if;
					end if;
				
				when ntp_polling =>
					if ntp_frame_rcvd_s='1' then
						no_ntp_timeout<='1';
					elsif timer_enable='1' and ntp_polling_cnt=0 then
						no_ntp_timeout<='0';
					end if;

					if timer_enable='1' then

						if ntp_polling_cnt=0 then
							do_ntp<='1';
							ntp_polling_cnt<=2**poll_interval_exp-1;

							if no_ntp_timeout='0' then
								if ntp_tmout_cnt=3 then
									ntp_timeout<='1';
									ntp_tmout_cnt<=0;
								else
									ntp_tmout_cnt<=ntp_tmout_cnt+1;
								end if;
							else
								ntp_tmout_cnt<=0;
							end if;
						else
							ntp_polling_cnt<=ntp_polling_cnt-1;
						end if;
					else
						ntp_timeout<='0';
						do_ntp<='0';
					end if;
			end case;
		end if;
	end if;
end process;

-------------------------------------------------------------------------------
-- local_time synchronizer
-------------------------------------------------------------------------------
process(mac_clk)
begin
	if rising_edge(mac_clk) then
		new_time_avail_1ck<=new_time_avail;
		new_time_avail_2ck<=new_time_avail_1ck;
		if new_time_avail_2ck='0' and new_time_avail_1ck='1' then
			local_time_sync<=local_time;
			reference_time_sync<=reference_time; -- it is updated on new_theta, but OK to use new_time_avail
		end if;

		do_ntp_1ck<=do_ntp;
		do_ntp_2ck<=do_ntp_1ck;
		if do_ntp_2ck='0' and do_ntp_1ck='1' then
			do_request_ntp<='1';
		else
			do_request_ntp<='0';
		end if;

	end if;
end process;

process(refclk)
begin
	if rising_edge(refclk) then
		rcv_ntp_ack_stretch_1ck<=rcv_ntp_ack_stretch;
		rcv_ntp_ack_stretch_2ck<=rcv_ntp_ack_stretch_1ck;
		if rcv_ntp_ack_stretch_2ck='0' and rcv_ntp_ack_stretch_1ck='1' then
			ntp_frame_rcvd_s<='1';
		else
			ntp_frame_rcvd_s<='0';
		end if;
			
		ntp_frame_rcvd<=ntp_frame_rcvd_s;
	end if;
end process;

-------------------------------------------------------------------------------
-- NTP TX
-------------------------------------------------------------------------------
lbl_udp_tx : ntp_udp_tx generic map(
	stratum=>stratum,
	poll_interval_exp=>poll_interval_exp,
	clock_precision=>clock_precision
	)
port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	our_ip_address=>our_ip_address,
	ntp_server_add=>ntp_server_add,

	do_request_ntp=>do_request_ntp,
	origin_time=>X"0000_0000_0000_0000",
	receive_time=>X"0000_0000_0000_0000",
	transmit_time=>local_time_sync,
	reference_time=>reference_time_sync,

	-- UDP TX signals
	udp_tx_start=>udp_tx_start,
	udp_txi=>udp_txi,
	udp_tx_result=>udp_tx_result,
	udp_tx_data_out_ready=>udp_tx_data_out_ready,

	-- UDP TX signals (from next stage)
	udp_tx_start_prior=>udp_tx_start_prior,
	udp_txi_prior=>udp_txi_prior,
	udp_tx_result_prior=>udp_tx_result_prior,
	udp_tx_data_out_ready_prior=>udp_tx_data_out_ready_prior
	);
-------------------------------------------------------------------------------
-- ntp RX state machine
-------------------------------------------------------------------------------
lbl_ntp_udp_rx : ntp_udp_rx port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	-- UDP RX signals
    udp_rx_start=>udp_rx_start,
    udp_rxo=>udp_rxo,

	local_time=>local_time_sync,
	local_time_at_rcv=>T_2,

	rcv_ntp_ack=>rcv_ntp_ack,
	origin_time=>T_1,
	receive_time=>T_prime_1,
	Transmit_time=>T_prime_2
	);

end RTL;
