library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity ntp_udp_server_sim is 
generic(
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_client_add 	: in std_logic_vector (31 downto 0);

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic;

	-- refclk interface (xtime slower than mac_clk)
	refclk: in std_logic
	);
end ntp_udp_server_sim;

architecture RTL of ntp_udp_server_sim is
component ntp_udp_rx
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	-- UDP RX signals
    udp_rx_start : in std_logic;       -- indicates receipt of udp header
    udp_rxo      : in udp_rx_type;

	local_time : in std_logic_vector(63 downto 0); -- local time
	local_time_at_rcv : out std_logic_vector(63 downto 0); -- local time at datagram rcv time

	rcv_ntp_ack: out std_logic;
	origin_time : out std_logic_vector(63 downto 0); -- rcvd OT field
	receive_time : out std_logic_vector(63 downto 0); -- rcvd RT field
	Transmit_time : out std_logic_vector(63 downto 0) -- rcvd TT field
	);
end component ntp_udp_rx;

component ntp_udp_tx
generic(
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);

	do_request_ntp : in std_logic;
	origin_time : in std_logic_vector(63 downto 0);
	receive_time : in std_logic_vector(63 downto 0);
	transmit_time : in std_logic_vector(63 downto 0); -- local time (hence transmit time)
	reference_time : in std_logic_vector(63 downto 0); -- last time when system was updated

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic
	);
end component ntp_udp_tx;

constant time_incr : unsigned(63 downto 0):=to_unsigned(2**(32-clock_precision),64);

signal local_time : std_logic_vector(63 downto 0):=X"0000_0001_0000_0000";
-- signal local_time : std_logic_vector(63 downto 0):=X"0000_0000_0000_0000";
-- signal local_time : std_logic_vector(63 downto 0):=X"D8E9_E028_7E29_0641";
signal rcv_ntp_ack: std_logic;
signal do_request_ntp: std_logic;
	
signal rx_origin_time : std_logic_vector(63 downto 0);
signal rx_receive_time : std_logic_vector(63 downto 0);
signal rx_Transmit_time : std_logic_vector(63 downto 0);
signal local_time_at_rcv : std_logic_vector(63 downto 0);

signal tx_origin_time : std_logic_vector(63 downto 0);
signal tx_receive_time : std_logic_vector(63 downto 0);
signal tx_Transmit_time : std_logic_vector(63 downto 0);

signal rand : natural:=0;
begin
-------------------------------------------------------------------------------
-- ntp RX state machine
-------------------------------------------------------------------------------
lbl_ntp_udp_rx : ntp_udp_rx port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	-- UDP RX signals
    udp_rx_start=>udp_rx_start,
    udp_rxo=>udp_rxo,

	local_time=>local_time,
	local_time_at_rcv=>local_time_at_rcv,

	rcv_ntp_ack=>rcv_ntp_ack,
	origin_time=>rx_origin_time,
	receive_time=>rx_receive_time,
	Transmit_time=>rx_Transmit_time
	);
-------------------------------------------------------------------------------
-- LOCAL TIME
-------------------------------------------------------------------------------
process
begin
-- 	local_time<=std_logic_vector(unsigned(local_time)+X"0000_0000_0000_0010");
	local_time<=std_logic_vector(unsigned(local_time)+time_incr);
	wait until rising_edge(refclk);
end process;
-------------------------------------------------------------------------------
-- COMPUTE AND GENERATE RESPONSE
-------------------------------------------------------------------------------
process
begin
	do_request_ntp<='0';
	wait until rising_edge(mac_clk) and rcv_ntp_ack='1';
	wait until rising_edge(mac_clk) and rcv_ntp_ack='0';
	
	tx_origin_time<=rx_transmit_time;
	tx_receive_time<=local_time_at_rcv;
	case rand is
-- 		when 0 =>
-- 			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0002_5000_0000");
-- 			rand<=rand+1;
-- 		when 1 => 
-- 			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0001_2000_0000");
-- 			rand<=rand+1;
-- 		when 2 => 
-- 			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0002_2000_0000");
-- 			rand<=rand+1;
-- 		when others => 
-- 			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0000_8000_0000");
-- 			rand<=0;
		when 0 =>
			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0000_0025_0000");
			rand<=rand+1;
		when 1 => 
			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0000_0012_0000");
			rand<=rand+1;
		when 2 => 
			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0000_0022_0000");
			rand<=rand+1;
		when others => 
			tx_transmit_time<=std_logic_vector(unsigned(local_time_at_rcv)+X"0000_0000_0080_0000");
			rand<=0;
	end case;
	wait until rising_edge(mac_clk);

	do_request_ntp<='1';
	wait until rising_edge(mac_clk);
	
	do_request_ntp<='0';
	wait until rising_edge(mac_clk);
end process;

-------------------------------------------------------------------------------
-- NTP TX
-------------------------------------------------------------------------------
lbl_udp_tx : ntp_udp_tx generic map(
	stratum=>7,
	poll_interval_exp=>poll_interval_exp,
	clock_precision=>clock_precision
	)
port map(
	mac_clk=>mac_clk,
	rst_macclk=>rst_macclk,

	our_ip_address=>our_ip_address,
	ntp_server_add=>ntp_client_add,

	do_request_ntp=>do_request_ntp,
	origin_time=>tx_origin_time,
	receive_time=>tx_receive_time,
	transmit_time=>tx_transmit_time,
	reference_time=>X"0000_0000_0000_0000",

	-- UDP TX signals
	udp_tx_start=>udp_tx_start,
	udp_txi=>udp_txi,
	udp_tx_result=>udp_tx_result,
	udp_tx_data_out_ready=>udp_tx_data_out_ready,

	-- UDP TX signals (from next stage)
	udp_tx_start_prior=>udp_tx_start_prior,
	udp_txi_prior=>udp_txi_prior,
	udp_tx_result_prior=>udp_tx_result_prior,
	udp_tx_data_out_ready_prior=>udp_tx_data_out_ready_prior
	);

end RTL;
