-- ntp_udp_tx
--
-- manage the communication with the UDP layer.
-- Use local time and provides T'1,T'2, T1 and T2
--
-- poll interval set to 128s (exponent of 7)
-- timeout value is equal to the poll interval
--
-- Olivier Bourrion, May 2015

library ieee;
use ieee.std_logic_1164.all;
use IEEE.NUMERIC_STD.ALL;

use work.pack_ipv4_types.all;
use work.pack_axi.all;
use work.pack_arp_types.all;

entity ntp_udp_tx is 
generic(
	stratum :  natural:=7; -- Must be >1, 6 or above recommanded
	poll_interval_exp :  integer:=7; -- 7 for 2^7=128s -- should be between 6 and 10
	clock_precision :  integer:=20 -- defined as 1s/2^-clock precision (20 for ~ 1 us)
	);
port(
	mac_clk: in std_logic; -- Ethernet MAC clock (125MHz)
	rst_macclk: in std_logic; -- MAC clock domain sync reset

	our_ip_address 	: in std_logic_vector (31 downto 0);
	ntp_server_add 	: in std_logic_vector (31 downto 0);

	do_request_ntp : in std_logic;
	origin_time : in std_logic_vector(63 downto 0);
	receive_time : in std_logic_vector(63 downto 0);
	transmit_time : in std_logic_vector(63 downto 0); -- local time (hence transmit time)
	reference_time : in std_logic_vector(63 downto 0); -- last time when system was updated

	-- UDP TX signals
	udp_tx_start			: out std_logic;					-- indicates req to tx UDP
	udp_txi					: out udp_tx_type;					-- UDP tx cxns
	udp_tx_result			: in std_logic_vector (1 downto 0); -- tx status (changes during transmission)
	udp_tx_data_out_ready   : in std_logic;						-- indicates udp_tx is ready to take data

	-- UDP TX signals (from next stage)
	udp_tx_start_prior	    : in std_logic;
	udp_txi_prior			: in udp_tx_type;
	udp_tx_result_prior		: out std_logic_vector (1 downto 0);
	udp_tx_data_out_ready_prior: out std_logic
	);
end ntp_udp_tx;

architecture RTL of ntp_udp_tx is
-- TX side
type def_state_ntp_tx is (idle,set_len,xmit_ntp,xmit_ntp_done);
signal state_ntp_tx : def_state_ntp_tx;
signal saved_transmit_time : std_logic_vector(63 downto 0);
signal tx_count : natural range 0 to 47;

-- UDP common
signal udp_tx_result_ntp		: std_logic_vector (1 downto 0);
signal udp_tx_data_out_ready_ntp: std_logic;

-- arbitrer
constant nb_src : natural :=2;
signal src_udp_tx_start : std_logic_vector(nb_src-1 downto 0);
signal src_udp_tx : udp_tx_vect_type(nb_src-1 downto 0);

type def_state_arb is (idle,busy);
signal state_arb_udp : def_state_arb;

signal sel : natural range 0 to nb_src-1;
begin
-------------------------------------------------------------------------------
-- NTP TX state machine
-------------------------------------------------------------------------------
src_udp_tx(1).hdr.src_port<=std_logic_vector(to_unsigned(123,16));
src_udp_tx(1).hdr.checksum<=(others=>'0');
src_udp_tx(1).hdr.dst_port<=std_logic_vector(to_unsigned(123,16));
src_udp_tx(1).hdr.dst_ip_addr<=ntp_server_add;
src_udp_tx(1).hdr.data_length<=std_logic_vector(to_unsigned(48,16));

process(mac_clk)
begin
	if rising_edge(mac_clk) then
		if rst_macclk='1' then
			src_udp_tx_start(1)<='0';
			src_udp_tx(1).data.data_out_valid<='0';
			src_udp_tx(1).data.data_out_last<='0';
			src_udp_tx(1).data.data_out<=(others=>'0');
			
			state_ntp_tx<=idle;
		else
			case state_ntp_tx is
				when idle =>
					tx_count<=0;
					if do_request_ntp='1' then
						state_ntp_tx<=set_len;
					end if;

				when set_len =>
					saved_transmit_time<=transmit_time;
					src_udp_tx_start(1)<='1';
					src_udp_tx(1).data.data_out_valid<='1';
					-- LI=0 (Leap indicator nNOWARN), VN=4 (NTP version), Mode=3 (client)
					src_udp_tx(1).data.data_out<=X"23";
					tx_count<=tx_count+1;
					state_ntp_tx<=xmit_ntp;

				when xmit_ntp =>
					if udp_tx_result_ntp = UDPTX_RESULT_ERR then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						state_ntp_tx<=IDLE;

					elsif udp_tx_data_out_ready_ntp='1' then
						if tx_count=47 then
							tx_count<=0;
							state_ntp_tx<=xmit_ntp_done;
						else
							tx_count<=tx_count+1;
						end if;

						case tx_count is
							when 1  => src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(stratum,8)); 
							when 2  => src_udp_tx(1).data.data_out<=std_logic_vector(to_unsigned(poll_interval_exp,8)); -- Polling interval
							when 3  => src_udp_tx(1).data.data_out<=std_logic_vector(to_signed(0-clock_precision,8)); -- Precision of local clock
							
							when 4  => src_udp_tx(1).data.data_out<=X"00"; -- Root delay (31..24) -- dummy value of 1s
							when 5  => src_udp_tx(1).data.data_out<=X"01"; -- Root delay (23..16)
							when 6  => src_udp_tx(1).data.data_out<=X"00"; -- Root delay (15..8)
							when 7  => src_udp_tx(1).data.data_out<=X"00"; -- Root delay (7..0)
							
							when 8  =>  src_udp_tx(1).data.data_out<=X"00"; -- Root dispersion (31..24) -- dummy value of 1s
							when 9  =>  src_udp_tx(1).data.data_out<=X"01"; -- Root dispersion (23..16)
							when 10  => src_udp_tx(1).data.data_out<=X"00"; -- Root dispersion (15..8)
							when 11  => src_udp_tx(1).data.data_out<=X"00"; -- Root dispersion (7..0)
							
							when 12  => src_udp_tx(1).data.data_out<=ntp_server_add(31 downto 24); -- Reference ID (IP for stratum>1)
							when 13  => src_udp_tx(1).data.data_out<=ntp_server_add(23 downto 16); -- Reference ID (IP for stratum>1)
							when 14  => src_udp_tx(1).data.data_out<=ntp_server_add(15 downto 8); -- Reference ID (IP for stratum>1)
							when 15  => src_udp_tx(1).data.data_out<=ntp_server_add(7 downto 0); -- Reference ID (IP for stratum>1)

							when 16  => src_udp_tx(1).data.data_out<=reference_time(63 downto 56); -- Reference time stamp (63..56) (NULL)
							when 17  => src_udp_tx(1).data.data_out<=reference_time(55 downto 48); -- Reference time stamp (55..48) (NULL)
							when 18  => src_udp_tx(1).data.data_out<=reference_time(47 downto 40); -- Reference time stamp (47..40) (NULL)
							when 19  => src_udp_tx(1).data.data_out<=reference_time(39 downto 32); -- Reference time stamp (39..32) (NULL)

							when 20  => src_udp_tx(1).data.data_out<=reference_time(31 downto 24); -- Reference time stamp (31 ..24) (NULL)
							when 21  => src_udp_tx(1).data.data_out<=reference_time(23 downto 16); -- Reference time stamp (23..16)(NULL)
							when 22  => src_udp_tx(1).data.data_out<=reference_time(15 downto 8 ); -- Reference time stamp (15..8) (NULL)
							when 23  => src_udp_tx(1).data.data_out<=reference_time( 7 downto 0 ); -- Reference time stamp (7..0) (NULL)

							when 24  => src_udp_tx(1).data.data_out<=origin_time(63 downto 56); -- Origin time stamp (63..56) (NULL)
							when 25  => src_udp_tx(1).data.data_out<=origin_time(55 downto 48); -- Origin time stamp (55..48) (NULL)
							when 26  => src_udp_tx(1).data.data_out<=origin_time(47 downto 40); -- Origin time stamp (47..40) (NULL)
							when 27  => src_udp_tx(1).data.data_out<=origin_time(39 downto 32); -- Origin time stamp (39..32) (NULL)

							when 28  => src_udp_tx(1).data.data_out<=origin_time(31 downto 24); -- Origin time stamp (31 ..24) (NULL)
							when 29  => src_udp_tx(1).data.data_out<=origin_time(23 downto 16); -- Origin time stamp (23..16)(NULL)
							when 30  => src_udp_tx(1).data.data_out<=origin_time(15 downto 8 ); -- Origin time stamp (15..8) (NULL)
							when 31  => src_udp_tx(1).data.data_out<=origin_time( 7 downto 0 ); -- Origin time stamp (7..0) (NULL)

							when 32  => src_udp_tx(1).data.data_out<=receive_time(63 downto 56); -- Receive time stamp (63..56) (NULL)
							when 33  => src_udp_tx(1).data.data_out<=receive_time(55 downto 48); -- Receive time stamp (55..48) (NULL)
							when 34  => src_udp_tx(1).data.data_out<=receive_time(47 downto 40); -- Receive time stamp (47..40) (NULL)
							when 35  => src_udp_tx(1).data.data_out<=receive_time(39 downto 32); -- Receive time stamp (39..32) (NULL)
                                                                                              
							when 36  => src_udp_tx(1).data.data_out<=receive_time(31 downto 24); -- Receive time stamp (31 ..24) (NULL)
							when 37  => src_udp_tx(1).data.data_out<=receive_time(23 downto 16); -- Receive time stamp (23..16)(NULL)
							when 38  => src_udp_tx(1).data.data_out<=receive_time(15 downto 8 ); -- Receive time stamp (15..8) (NULL)
							when 39  => src_udp_tx(1).data.data_out<=receive_time( 7 downto 0 ); -- Receive time stamp (7..0) (NULL)

							when 40  => src_udp_tx(1).data.data_out<=saved_transmit_time(63 downto 56); -- Transmit time stamp (63..56) 
							when 41  => src_udp_tx(1).data.data_out<=saved_transmit_time(55 downto 48); -- Transmit time stamp (55..48) 
							when 42  => src_udp_tx(1).data.data_out<=saved_transmit_time(47 downto 40); -- Transmit time stamp (47..40) 
							when 43  => src_udp_tx(1).data.data_out<=saved_transmit_time(39 downto 32); -- Transmit time stamp (39..32) 

							when 44  => src_udp_tx(1).data.data_out<=saved_transmit_time(31 downto 24); -- Transmit time stamp (31 ..24) 
							when 45  => src_udp_tx(1).data.data_out<=saved_transmit_time(23 downto 16); -- Transmit time stamp (23..16)
							when 46  => src_udp_tx(1).data.data_out<=saved_transmit_time(15 downto 8); -- Transmit time stamp (15..8) 
							when 47  => 
									src_udp_tx(1).data.data_out<=saved_transmit_time(7 downto 0); -- Transmit time stamp (7..0) 
									src_udp_tx(1).data.data_out_last<='1';

							when others  => NULL;
						end case;
					end if;
				
				when xmit_ntp_done =>
					if udp_tx_result_ntp = UDPTX_RESULT_ERR then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						src_udp_tx_start(1)<='0';
						state_ntp_tx<=IDLE;

					elsif udp_tx_data_out_ready_ntp='1' then
						src_udp_tx(1).data.data_out_valid<='0';
						src_udp_tx(1).data.data_out_last<='0';
						src_udp_tx(1).data.data_out<=(others=>'0');
						src_udp_tx_start(1)<='0';
						state_ntp_tx<=IDLE;
					end if;

			end case;
		end if;
	end if;
end process;
-------------------------------------------------------------------------------
-- UDP arbitrer
-------------------------------------------------------------------------------
src_udp_tx_start(0)<=udp_tx_start_prior;
src_udp_tx(0)<=udp_txi_prior;

process(mac_clk)
begin
	if rising_edge(mac_clk) then
		if rst_macclk = '1' then
			sel<=0;
			state_arb_udp<=idle;
		else
			case state_arb_udp is
				when idle =>
					if src_udp_tx_start(sel)='1' then
						state_arb_udp<=busy;
					elsif sel=nb_src-1 then
						sel<=0;
					else
						sel<=sel+1;
					end if;

				when busy =>
					if src_udp_tx(sel).data.data_out_last='1' then
						state_arb_udp<=idle;
					end if;
			end case;
		end if;
	end if;
end process;

udp_tx_result_prior<=udp_tx_result;
udp_tx_result_ntp<=udp_tx_result;

udp_tx_data_out_ready_prior<=udp_tx_data_out_ready when sel=0 else '0';
udp_tx_data_out_ready_ntp <=udp_tx_data_out_ready when sel=1 else '0';

udp_tx_start<=src_udp_tx_start(sel);
udp_txi <= src_udp_tx(sel);

end RTL;
