----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 		Peter Fall, rewrite by Olivier Bourrion (fusion of process, sync between tx and RX)
-- 
-- Create Date:    16:20:42 06/01/2011 
-- Design Name: 
-- Module Name:    IPv4_RX - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--              handle simple IP RX
--              doesnt handle reassembly
--              checks and filters for IP protocol
--              checks and filters for IP addr
--              Handle IPv4 protocol
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Revision 0.02 - Improved error handling
-- Revision 0.03 - Added handling of broadcast address
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use work.pack_axi.all;
use work.pack_ipv4_types.all;
use work.pack_arp_types.all;

entity IPv4_RX is
  port (
    -- IP Layer signals
    ip_rx             : out ipv4_rx_type;
    ip_rx_start       : out std_logic;  -- indicates receipt of ip frame.
    -- system signals
    clk               : in  std_logic;  -- same clock used to clock mac data and ip data
    reset             : in  std_logic;
    our_ip_address    : in  std_logic_vector (31 downto 0);
	our_mac_address 		: in std_logic_vector (47 downto 0);
    rx_pkt_count      : out std_logic_vector(7 downto 0);   -- number of IP pkts received for us
    -- MAC layer RX signals
    mac_data_in       : in  std_logic_vector (7 downto 0);  -- ethernet frame (from dst mac addr through to last byte of frame)
    mac_data_in_valid : in  std_logic;  -- indicates data_in valid on clock
	mac_rx_terror        : in  std_logic;							-- tells mac produced an error
    mac_data_in_last  : in  std_logic   -- indicates last data in frame
    );                  
end IPv4_RX;

architecture Behavioral of IPv4_RX is
type rx_state_ipv4_type is (IDLE, ETH_HDR, IP_HDR, USER_DATA, WAIT_END);

-- state variables
signal rx_state_ipv4         : rx_state_ipv4_type;
signal rx_count         : unsigned (15 downto 0);
signal data_len         : std_logic_vector (15 downto 0);  -- src data length captured from input
signal frame_err_cnt    : unsigned (7 downto 0);  -- number of frame errors
signal rx_pkt_counter   : unsigned (7 downto 0);  -- number of rx frames received for us
signal ip_is_allocated : std_logic;
-- rx control signals
signal dst_ip_rx         : std_logic_vector (31 downto 0);

-- IP datagram header format
--
--      0          4          8                      16      19             24                    31
--      --------------------------------------------------------------------------------------------
--      | Version  | *Header  |    Service Type      |        Total Length including header        |
--      |   (4)    |  Length  |     (ignored)        |                 (in bytes)                  |
--      --------------------------------------------------------------------------------------------
--      |           Identification                   | Flags |       Fragment Offset               |
--      |                                            |       |      (in 32 bit words)              |
--      --------------------------------------------------------------------------------------------
--      |    Time To Live     |       Protocol       |             Header Checksum                 |
--      |     (ignored)       |                      |                                             |
--      --------------------------------------------------------------------------------------------
--      |                                   Source IP Address                                      |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
--      |                                 Destination IP Address                                   |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
--      |                          Options (if any - ignored)               |       Padding        |
--      |                                                                   |      (if needed)     |
--      --------------------------------------------------------------------------------------------
--      |                                          Data                                            |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
--      |                                          ....                                            |
--      |                                                                                          |
--      --------------------------------------------------------------------------------------------
--
-- * - in 32 bit words 
  
begin
-------------------------------------------------------------------------------
-- process IP frame
-------------------------------------------------------------------------------
ip_rx.hdr.num_frame_errors <= std_logic_vector(frame_err_cnt);
rx_pkt_count               <= std_logic_vector(rx_pkt_counter);
ip_rx.hdr.data_length      <= data_len;

rx_sequential : process (clk)
begin
	if rising_edge(clk) then
		if reset = '1' then
			rx_state_ipv4     <= IDLE;
			dst_ip_rx         <= (others => '0');
			rx_count         <= x"0000";
			ip_rx.hdr.src_ip_addr<= (others => '0');
			ip_rx.hdr.protocol<= (others => '0');
			ip_rx.hdr.is_valid   <= '0';
			ip_rx.hdr.last_error_code<= RX_EC_NONE;
			data_len         <= (others => '0');
			ip_rx_start  <= '0';
			frame_err_cnt    <= (others => '0');
			rx_pkt_counter   <= x"00";
		else
			if our_ip_address=X"00_00_00_00" then
				ip_is_allocated<='0';
			else
				ip_is_allocated<='1';
			end if;

			-- RX FSM
			case rx_state_ipv4 is
				when IDLE =>
					ip_rx.hdr.is_broadcast   <= '0';
					ip_rx.hdr.is_valid   <= '0';
					ip_rx.data.data_in       <= (others => '0');
					ip_rx.data.data_in_valid <= '0';
					ip_rx.data.data_in_last  <= '0';
					ip_rx.hdr.last_error_code<= RX_EC_NONE;
					if mac_data_in_valid = '1' then
						rx_count <= rx_count + 1;
						rx_state_ipv4 <= ETH_HDR;
					else
						rx_count <= (others=>'0');
					end if;

				when ETH_HDR =>
					if mac_data_in_valid = '1' then
						if rx_count = x"000d" then
							rx_count <= (others=>'0');
							rx_state_ipv4 <= IP_HDR;
						else
							rx_count <= rx_count + 1;
						end if;
															-- handle early frame termination
						if mac_data_in_last = '1' then
							rx_count <= (others=>'0');
							ip_rx.hdr.last_error_code<= RX_EC_ET_ETH;
							frame_err_cnt <= frame_err_cnt + 1;
							ip_rx_start <= '0';
							rx_state_ipv4     <= IDLE;
						else
							case rx_count is
								when x"0000" =>
									if mac_data_in /= our_mac_address(47 downto 40) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"0001" =>
									if mac_data_in /= our_mac_address(39 downto 32) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"0002" =>
									if mac_data_in /= our_mac_address(31 downto 24) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"0003" =>
									if mac_data_in /= our_mac_address(23 downto 16) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"0004" =>
									if mac_data_in /= our_mac_address(15 downto 8) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"0005" =>
									if mac_data_in /= our_mac_address(7 downto 0) and mac_data_in /=x"FF" then  -- ignore pkts that are not bc or our mac
										rx_state_ipv4 <= WAIT_END;
									end if;

								when x"000c" =>
									if mac_data_in /= x"08" then  -- ignore pkts that are not type=IP
										rx_state_ipv4 <= WAIT_END;
									end if;
								  
								when x"000d" =>
									if mac_data_in /= x"00" then  -- ignore pkts that are not type=IP
										rx_state_ipv4 <= WAIT_END;
									end if;
								  
								when others =>          -- ignore other bytes in eth header
							end case;
						end if;
					end if;

				when IP_HDR =>
					if mac_data_in_valid = '1' then
						-- handle early frame termination
						if mac_data_in_last = '1' then
							ip_rx.hdr.last_error_code<= RX_EC_ET_IP;
							frame_err_cnt <= frame_err_cnt + 1;
							ip_rx_start <= '0';
							rx_count <= (others=>'0');
							rx_state_ipv4     <= IDLE;
						else
							if (rx_count/=x"0013") then 		
								rx_count <= rx_count+1;
							else
								rx_count <= x"0001";
							end if;

							case rx_count is
								when x"0000" =>
								  if mac_data_in /= x"45" then  -- ignore pkts that are not v4 with 5 header words
									rx_state_ipv4 <= WAIT_END;
								  end if;
								  
								when x"0002" => 
									  data_len (15 downto 8) <= mac_data_in;
									  data_len (7 downto 0)  <= x"00";
								when x"0003" => 
									-- compute data length, taking into account that we need to subtract the header length
									data_len <= std_logic_vector(unsigned(data_len(15 downto 8) & mac_data_in) - 20); -- subtract header length

								when x"0006" =>
									if (mac_data_in(7) = '1') or (mac_data_in (4 downto 0) /= "00000") then
										-- ignore pkts that require reassembly (MF=1 or frag offst /= 0)
										rx_state_ipv4 <= WAIT_END;
									end if;
								  
								when x"0007" =>
								  if mac_data_in /= x"00" then  -- ignore pkts that require reassembly (frag offst /= 0)
									rx_state_ipv4 <= WAIT_END;
								  end if;

								when x"0009" => ip_rx.hdr.protocol <= mac_data_in;

								when x"000c" => ip_rx.hdr.src_ip_addr(31 downto 24) <= mac_data_in; 
								when x"000d" => ip_rx.hdr.src_ip_addr(23 downto 16) <= mac_data_in; 
								when x"000e" => ip_rx.hdr.src_ip_addr(15 downto 8) <= mac_data_in; 
								when x"000f" => ip_rx.hdr.src_ip_addr(7 downto 0) <= mac_data_in; 

								when x"0010" => 
									dst_ip_rx(31 downto 24) <= mac_data_in;
									if ((mac_data_in /= our_ip_address(31 downto 24)) and (mac_data_in /= IP_BC_ADDR(31 downto 24)) and ip_is_allocated='1')then  -- ignore pkts that are not addressed to us
										rx_state_ipv4 <= WAIT_END;
									end if;
								when x"0011" => 
									dst_ip_rx(23 downto 16) <= mac_data_in;
									if ((mac_data_in /= our_ip_address(23 downto 16)) and (mac_data_in /= IP_BC_ADDR(23 downto 16)) and ip_is_allocated='1')then  -- ignore pkts that are not addressed to us
										rx_state_ipv4 <= WAIT_END;
									  end if;
								when x"0012" => 
									dst_ip_rx(15 downto 8) <= mac_data_in;
									if ((mac_data_in /= our_ip_address(15 downto 8)) and (mac_data_in /= IP_BC_ADDR(15 downto 8)) and ip_is_allocated='1')then  -- ignore pkts that are not addressed to us
										rx_state_ipv4 <= WAIT_END;
									  end if;

								when x"0013" =>
									dst_ip_rx(7 downto 0) <= mac_data_in;
									if ((mac_data_in /= our_ip_address(7 downto 0)) and (mac_data_in /= IP_BC_ADDR(7 downto 0)) and ip_is_allocated='1')then  -- ignore pkts that are not addressed to us
										rx_state_ipv4 <= WAIT_END;
									else
										rx_state_ipv4   <= USER_DATA;
										rx_pkt_counter <= rx_pkt_counter + 1;
										ip_rx_start <= '1';
									end if;

									-- now have the dst IP addr
									if (dst_ip_rx(31 downto 8) = IP_BC_ADDR(31 downto 8) and mac_data_in = IP_BC_ADDR(7 downto 0)) then
										ip_rx.hdr.is_broadcast     <= '1';
									else
										ip_rx.hdr.is_broadcast     <= '0';
									end if;
									ip_rx.hdr.is_valid<= '1';  -- header values are now valid, although the pkt may not be for us

								when others =>  -- ignore other bytes in ip header                                                                               
						  end case;
						end if;
					end if;
				
				when USER_DATA =>
					ip_rx.data.data_in       <= mac_data_in;
					ip_rx.data.data_in_valid <= mac_data_in_valid;

					if mac_data_in_valid = '1' then
						-- note: data gets transfered upstream as part of "output followers" processing
						if rx_count = unsigned(data_len) then
							rx_count <= x"0000";
							ip_rx_start<= '0';
							ip_rx.data.data_in_last  <= '1';
							if mac_data_in_last = '1' then
								if mac_rx_terror= '1' then
									ip_rx.hdr.last_error_code<= RX_EC_CRC;
								end if;
								rx_state_ipv4   <= IDLE;
							else
								rx_state_ipv4 <= WAIT_END;
							end if;
						else
							-- check for early frame termination
							if mac_data_in_last = '1' then
								ip_rx.data.data_in_last  <= '1';
								ip_rx.hdr.last_error_code<= RX_EC_ET_USER;
								frame_err_cnt <= frame_err_cnt + 1;
								ip_rx_start <= '0';
								rx_count <= x"0000";
								rx_state_ipv4     <= IDLE;
							else
								rx_count <= rx_count+1;
							end if;
						end if;
					end if;

				  when WAIT_END =>
					ip_rx.data.data_in_last  <= '0';
					ip_rx.data.data_in       <= x"00";
					ip_rx.data.data_in_valid <= '0';
					if mac_data_in_valid = '1' then
						if mac_data_in_last = '1' then
							if mac_rx_terror= '1' then
								ip_rx.hdr.last_error_code<= RX_EC_CRC;
							end if;
							rx_state_ipv4   <= IDLE;
						end if;
					end if;
				end case;
			end if;
		end if;
end process;
end Behavioral;

