----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 		Peter Fall
-- 
-- Create Date:    5 June 2011 
-- Design Name: 
-- Engineer: 		Peter Fall, rewrite by Olivier Bourrion (fusion of process, sync between tx and RX)
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--		handle simple UDP TX
--		doesnt generate the checksum(supposedly optional)
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Revision 0.02 - Added abort of tx when receive last from upstream
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.pack_axi.all;
use work.pack_ipv4_types.all;

entity UDP_TX is
port (
	-- UDP Layer signals
	udp_tx_start			: in std_logic;							-- indicates req to tx UDP
	udp_txi					: in udp_tx_type;							-- UDP tx cxns
	udp_tx_result			: out std_logic_vector (1 downto 0);-- tx status (changes during transmission)
	udp_tx_data_out_ready: out std_logic;							-- indicates udp_tx is ready to take data
	-- system signals
	clk 						: in  STD_LOGIC;							-- same clock used to clock mac data and ip data
	reset 					: in  STD_LOGIC;

	-- arbitrator signals
	ip_layer_request : out std_logic;
	ip_layer_granted : in std_logic;

	-- IP layer TX signals
	ip_tx_start				: out std_logic;
	ip_tx					: out ipv4_tx_type;							-- IP tx cxns
	ip_tx_result			: in std_logic_vector (1 downto 0);		-- tx status (changes during transmission)
	ip_tx_data_out_ready	: in std_logic									-- indicates IP TX is ready to take data
	);
end UDP_TX;

architecture Behavioral of UDP_TX is
type tx_state_type is (IDLE,REQUEST,PAUSE, SEND_UDP_HDR, SEND_USER_DATA);
		
-- TX state variables
signal udp_tx_state		: tx_state_type;
signal tx_count 			: unsigned (15 downto 0);

-- tx temp signals
signal total_length		: std_logic_vector (15 downto 0);	-- computed combinatorially from header size
	
-- IP datagram header format
--
--	0          4          8                      16      19             24                    31
--	--------------------------------------------------------------------------------------------
--	|              source port number            |              dest port number               |
--	|                                            |                                             |
--	--------------------------------------------------------------------------------------------
--	|                length (bytes)              |                checksum                     |
--	|          (header and data combined)        |                                             |
--	--------------------------------------------------------------------------------------------
--	|                                          Data                                            |
--	|                                                                                          |
--	--------------------------------------------------------------------------------------------
--	|                                          ....                                            |
--	|                                                                                          |
--	--------------------------------------------------------------------------------------------
		
begin
ip_tx.hdr.data_length <= total_length;
ip_tx.hdr.dst_ip_addr <= udp_txi.hdr.dst_ip_addr;
ip_tx.hdr.protocol <= x"11";	-- UDP protocol

tx_sequential : process (clk)
begin
	if rising_edge(clk) then
		if reset = '1' then
			-- reset state variables
			udp_tx_state <= IDLE;
			tx_count <= x"0000";
			ip_tx_start <= '0';
			ip_tx.data.data_out_valid <='0';
			udp_tx_data_out_ready <= '0';
			ip_tx.data.data_out_last <= '0';
			ip_layer_request<='0';

		else
			total_length <= std_logic_vector(unsigned(udp_txi.hdr.data_length) + 8);		-- total length = user data length + header length (bytes)
			
			-- TX FSM
			case udp_tx_state is
				when IDLE =>
					udp_tx_data_out_ready <= '0';		-- in this state, we are unable to accept user data for tx
					ip_tx.data.data_out_valid <='0';
					ip_tx.data.data_out_last <= '0';
					tx_count <= x"0000";
					if udp_tx_start = '1' then
						-- check header count for error if too high
						if unsigned(udp_txi.hdr.data_length) > 1472 then
							udp_tx_result <=UDPTX_RESULT_ERR;
						else
							-- start to send UDP header
							udp_tx_result <=UDPTX_RESULT_SENDING;
							ip_layer_request <='1';
							udp_tx_state <= REQUEST;
						end if;
					end if;

				when REQUEST =>
					if ip_layer_granted='1' then
						ip_tx_start <= '1';
						udp_tx_state <= PAUSE;
					end if;

				when PAUSE =>
					-- delay one clock for IP layer to respond to ip_tx_start and remove any tx error result
					ip_tx.data.data_out_valid <='1';
					tx_count<= tx_count+1;
					ip_tx.data.data_out <= udp_txi.hdr.src_port (15 downto 8);	-- src port
					udp_tx_state <= SEND_UDP_HDR;
					
				when SEND_UDP_HDR =>
					udp_tx_data_out_ready <= '0';		-- in this state, we are unable to accept user data for tx
					if ip_tx_result = IPTX_RESULT_ERR then
						ip_tx_start <= '0';
						ip_tx.data.data_out_valid <='0';
						udp_tx_result <=UDPTX_RESULT_ERR;
						ip_layer_request <='0';
						udp_tx_state <= IDLE;
					elsif ip_tx_data_out_ready = '1' then
						if tx_count = x"0007" then
							tx_count<= x"0001";
							ip_tx.data.data_out<= udp_txi.data.data_out;
							udp_tx_data_out_ready <= ip_tx_data_out_ready;	-- in this state, we can accept user data if IP TX rdy
							udp_tx_state <= SEND_USER_DATA;
						else
							tx_count<= tx_count+1;
						end if;
						case tx_count is
							when x"0001"  => ip_tx.data.data_out <= udp_txi.hdr.src_port (7 downto 0);
							when x"0002"  => ip_tx.data.data_out <= udp_txi.hdr.dst_port (15 downto 8);	-- dst port
							when x"0003"  => ip_tx.data.data_out <= udp_txi.hdr.dst_port (7 downto 0);
							when x"0004"  => ip_tx.data.data_out <= total_length (15 downto 8);				-- length
							when x"0005"  => ip_tx.data.data_out <= total_length (7 downto 0);
							when x"0006"  => ip_tx.data.data_out <= udp_txi.hdr.checksum (15 downto 8);	-- checksum (set by upstream)
							when x"0007"  => ip_tx.data.data_out <= udp_txi.hdr.checksum (7 downto 0);
							when others =>
								-- shouldnt get here - handle as error
								udp_tx_result <= UDPTX_RESULT_ERR;
						end case;
					end if;
					
				when SEND_USER_DATA =>
					udp_tx_data_out_ready <= ip_tx_data_out_ready;	-- in this state, we can accept user data if IP TX rdy
					if ip_tx_data_out_ready = '1' then
-- 						if udp_txi.data.data_out_valid = '1' or tx_count = x"000" then
						if udp_txi.data.data_out_valid = '1' then
							-- only increment if ready and valid has been subsequently established, otherwise data count moves on too fast
							if unsigned(tx_count) = unsigned(udp_txi.hdr.data_length) then						
								-- TX terminated due to count - end normally
								ip_tx.data.data_out_last <= '1';
								ip_tx.data.data_out<= udp_txi.data.data_out;
								ip_tx_start <= '0';
								udp_tx_result <= UDPTX_RESULT_SENT;
								ip_layer_request <='0';
								udp_tx_state <= IDLE;
							elsif udp_txi.data.data_out_last = '1' then
								-- terminate tx with error as got last from upstream before exhausting count
								ip_tx.data.data_out_last <= '1';
								ip_tx.data.data_out<= udp_txi.data.data_out;
								ip_tx_start <= '0';
								udp_tx_result <= UDPTX_RESULT_ERR;
								ip_layer_request <='0';
								udp_tx_state <= IDLE;
							else
								-- TX continues
								tx_count<= tx_count+1;
								ip_tx.data.data_out<= udp_txi.data.data_out;
							end if;
						end if;
					end if;
				end case;
			end if;
		end if;
end process;
-- -------------------------------------------------------------------------------
-- -- checksum computation
-- -------------------------------------------------------------------------------
-- crc : process (clk_out)
-- begin
-- 	if rising_edge(clk_out) then
-- 		if (reset='1') then
-- 			crc_state  <= IDLE;
-- 		else
-- 			case crc_state is
-- 				when IDLE =>
-- 					if do_crc = '1' then
-- 						checksum<=x"00" & unsigned(our_ip_address(15 downto 0));
-- 						crc_state  <= SRC_IP_LSB;
-- 					end if;
-- 
-- 				when SRC_IP_LSB =>
-- 					checksum<=unsigned(checksum)+unsigned(our_ip_address(31 downto 16));
-- 					crc_state  <= SRC_IP_MSB;
-- 				
-- 				when SRC_IP_MSB =>
-- 					checksum<=unsigned(checksum)+unsigned(udp_txi.hdr.dst_ip_addr(15 downto 0));
-- 					crc_state  <= DST_IP_LSB;
-- 
-- 				when DST_IP_LSB =>
-- 					checksum<=unsigned(checksum)+unsigned(udp_txi.hdr.dst_ip_addr(31 downto 16));
-- 					crc_state  <= DST_IP_MSB;
-- 				
-- 				when USER_DATA =>
-- 					if (wr_ptr(0)='0') then -- every even number update checksum
-- 						checksum<=unsigned(checksum)+unsigned(data16);
-- 					end if;
-- 					if (do_reply='1') then
-- 						crc_state  <= ADDOVF;
-- 					end if;
-- 
-- 				when ADDOVF => -- 16 LSB + carry
-- 					checksum <= (unsigned(checksum) and x"00ffff")+ unsigned(checksum(23 downto 16));
-- 					crc_state  <= FINAL;
-- 
-- 				when FINAL => -- flip every bit
-- 					checksum <= not(checksum);
-- 					crc_state  <= IDLE;
-- 			end case;
-- 		end if;
-- 	end if;
-- end process crc;


end Behavioral;
